# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

from gettext import gettext as _
from typing import Self, Any, override
import apsw
import numpy as np
import pandas as pd
from matplotlib.figure import Figure
from biomeca.plugin import register_loader, register_node, Node, BasicNode
from biomeca import c3d


@register_loader('c3d')
def load_c3d(f: apsw.Blob) -> pd.DataFrame:
    reader = c3d.Reader(f) # type: ignore
    frames = []
    point_data = []
    for frame, points, analog in reader.read_frames():
        frames.append(frame)
        point_data.append(points)
    point_cols = [(label.rstrip(), dim) for label in reader.point_labels for dim in 'XYZ']
    column_index = pd.MultiIndex.from_tuples(point_cols, names=['Label', 'Dimension'])
    data = np.empty((len(frames), len(point_cols)), dtype=float)
    for i, points in enumerate(point_data):
        data[i, :] = points[:, :3].flatten()
    return pd.DataFrame(data, index=frames, columns=column_index)


@register_loader('csv')
def load_csv(f: apsw.Blob) -> pd.DataFrame:
    return pd.read_csv(f) # type: ignore


@register_node
class GroupInput(Node):
    name = 'Group Input'

    @override
    @classmethod
    def create(cls, conn: apsw.Connection, nodegroup_id: int) -> Self:
        with conn:
            conn.execute('''
                insert into node (nodegroup_id, type) values (?, ?)
            ''', (nodegroup_id, cls.name))
            id = conn.last_insert_rowid()
            conn.execute('''
                insert into node_output (node_id, name, idx)
                    select ?, name, idx from nodegroup_input where nodegroup_id = ?
            ''', (id, nodegroup_id))
            # outputs kept up to date by a trigger
            return cls(conn, id)

    @override
    def run(self, inputs: list[Any]) -> list[Any]:
        raise Exception('The Group Input node should be handled specially by the runtime')


@register_node
class GroupOutput(Node):
    name = 'Group Output'

    @override
    @classmethod
    def create(cls, conn: apsw.Connection, nodegroup_id: int) -> Self:
        with conn:
            conn.execute('''
                insert into node (nodegroup_id, type) values (?, ?)
            ''', (nodegroup_id, cls.name))
            id = conn.last_insert_rowid()
            conn.execute('''
                insert into node_input (node_id, name, idx)
                    select ?, name, idx from nodegroup_output where nodegroup_id = ?
            ''', (id, nodegroup_id))
            # inputs kept up to date by a trigger
            return cls(conn, id)

    @override
    def run(self, inputs: list[Any]) -> list[Any]:
        raise Exception('The Group Output node should be handled specially by the runtime')


@register_node
class BoolInput(BasicNode):
    name = 'Boolean'
    inputs = [('Value', 'checkbox', None, 0)]
    outputs = ['Result']

    @override
    def run(self, inputs: list[Any]) -> list[Any]:
        return [bool(inputs[0])]


@register_node
class NumberInput(BasicNode):
    name = 'Number'
    inputs = [('Value', 'spinbox', None, 0.0)]
    outputs = ['Result']

    @override
    def run(self, inputs: list[Any]) -> list[Any]:
        return [float(inputs[0])]


@register_node
class StringInput(BasicNode):
    name = 'Text'
    inputs = [('Value', 'entry', None, '')]
    outputs = ['Result']

    @override
    def run(self, inputs: list[Any]) -> list[Any]:
        return [str(inputs[0])]


@register_node
class AddOperator(BasicNode):
    name = 'Add'
    inputs = [('Addend 1', 'spinbox', None, 0.0), ('Addend 2', 'spinbox', None, 0.0)]
    outputs = ['Result']

    @override
    def run(self, inputs: list[Any]) -> list[Any]:
        return [inputs[0] + inputs[1]]


@register_node
class SelectColumn(BasicNode):
    name = 'Select Column'
    inputs = [('Data', None, None, None), ('Level', 'combobox', {'values': []}, ''), ('Column', 'combobox', {'values': []}, '')]
    outputs = ['Data']

    @override
    def run(self, inputs: list[Any]) -> list[Any]:
        data, level, column = inputs
        if not isinstance(data, pd.DataFrame):
            return [None]

        is_multi = isinstance(data.columns, pd.MultiIndex)

        levels = data.columns.names if is_multi else ['default']
        if level not in levels:
            level = levels[0]
            self.set_input_value('Level', level)

        col_values = data.columns.get_level_values(level).values if is_multi else data.columns.values
        columns = list(set(col_values))
        if column not in columns:
            column = columns[0]
            self.set_input_value('Column', column)

        self.set_control_info('Level', {'values': list(levels)})
        self.set_control_info('Column', {'values': list(columns)})

        if not is_multi:
            level = None

        return [data.xs(column, level=level, axis='columns', drop_level=False)]


@register_node
class BarPlot(BasicNode):
    name = 'Bar Plot'
    inputs = [('Data', None, None, None), ('X column', 'combobox', {'values': []}, ''), ('Y column', 'combobox', {'values': []}, '')]
    outputs = ['Figure']

    @override
    def run(self, inputs: list[Any]) -> list[Any]:
        data, xcol, ycol = inputs
        if not isinstance(data, pd.DataFrame):
            return [None]

        columns = {(c if isinstance(c, str) else '.'.join(c)): c for c in data.columns.values}
        colnames = list(columns.keys())

        if xcol not in colnames:
            xcol = colnames[0]
            self.set_input_value('X column', xcol)
        if ycol not in colnames:
            ycol = colnames[1]
            self.set_input_value('Y column', ycol)

        self.set_control_info('X column', {'values': colnames})
        self.set_control_info('Y column', {'values': colnames})

        fig = Figure(figsize=(5, 4), dpi=100)
        axes = fig.add_subplot()
        axes.set_xlabel(xcol)
        axes.set_ylabel(ycol)
        x = list(range(len(data[columns[xcol]])))
        axes.bar(x, data[columns[ycol]], label=data[columns[xcol]])
        return [fig]


@register_node
class ScatterPlot(BasicNode):
    name = 'Scatter Plot'
    inputs = [('Data', None, None, None), ('X column', 'combobox', {'values': []}, ''), ('Y column', 'combobox', {'values': []}, '')]
    outputs = ['Figure']

    @override
    def run(self, inputs: list[Any]) -> list[Any]:
        data, xcol, ycol = inputs
        if not isinstance(data, pd.DataFrame):
            return [None]

        columns = {(c if isinstance(c, str) else '.'.join(c)): c for c in data.columns.values}
        colnames = ['index'] + list(columns.keys())

        if xcol not in colnames:
            xcol = 'index'
            self.set_input_value('X column', xcol)
        if ycol not in colnames:
            ycol = colnames[0]
            self.set_input_value('Y column', ycol)

        self.set_control_info('X column', {'values': colnames})
        self.set_control_info('Y column', {'values': colnames})

        fig = Figure(figsize=(5, 4), dpi=100)
        axes = fig.add_subplot()
        axes.set_xlabel(xcol)
        axes.set_ylabel(ycol)
        xdata = data.index.values if xcol == 'index' else data[columns[xcol]]
        ydata = data.index.values if ycol == 'index' else data[columns[ycol]]
        axes.plot(xdata, ydata)
        return [fig]
