# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import argparse
import datetime
import io
import logging
import sys
from gettext import gettext as _
from pathlib import Path
from typing import override
from biomeca import widgets, core


__version__ = '0.1.0'


logger = logging.getLogger(__name__)


class ISOFormatter(logging.Formatter):
    @override
    def formatTime(self, record: logging.LogRecord, datefmt: str | None = None) -> str:
        tz = datetime.datetime.now().astimezone().tzinfo
        dt = datetime.datetime.fromtimestamp(record.created, tz)
        return dt.isoformat(timespec='milliseconds')


def main(argv: list[str] = sys.argv) -> int:
    parser = argparse.ArgumentParser(prog='biomeca', description=_('Analyze biomechanics data'))
    parser.add_argument('filename', type=str, nargs='?', help=_('initial project file to open'))
    parser.add_argument('-V', '--version', action='store_true', help=_('print the program\'s version and exit'))
    parser.add_argument('-v', action='count', default=0, dest='verbose', help=_('produce more verbose output; increase the log level'))
    args = parser.parse_args(argv[1:])

    if args.version:
        print(f'biomeca v{__version__}')
        return 0

    if args.verbose == 0:
        level = logging.WARNING
    elif args.verbose == 1:
        level = logging.INFO
    elif args.verbose >= 2:
        level = logging.DEBUG

    root = logging.getLogger()
    root.setLevel(level)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(level)
    handler.setFormatter(ISOFormatter('{asctime} {levelname} {name}: {message}', style='{'))
    root.addHandler(handler)

    logger.info('Initializing')
    widgets.MainWindow(path=args.filename or ':memory:').root.mainloop()
    logger.info('Exiting with code 0')
    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))
