# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import json
import logging
import tkinter as tk
from functools import partial
from gettext import gettext as _
from tkinter import ttk
from typing import override, Callable, Any
import apsw
from biomeca import plugin
from biomeca.reactivity import ChangeListener
from biomeca.undo import undogroup


logger = logging.getLogger(__name__)


class NodeEditor(ttk.Frame):
    def __init__(self, parent: tk.Misc, conn: apsw.Connection, listener: ChangeListener) -> None:
        super().__init__(parent, relief='sunken', padding=5)
        self.conn = conn
        self.listener = listener

        self.id: int | None = None

        self.node_widgets: dict[int, NodeWidget] = {}

        self.setup()

        self.listener.register(self.on_selected_update, apsw.SQLITE_UPDATE, 'main', 'selected_nodegroup')
        self.listener.register(self.on_link_change, None, 'main', 'link')

    def setup(self) -> None:
        row = self.conn.execute('select nodegroup_id from selected_nodegroup').fetchone()
        assert row is not None
        id, = row
        if id == self.id:
            return
        self.id = id

        for widget in self.winfo_children():
            widget.destroy()

        self.node_widgets.clear()

        if self.id is None:
            ttk.Label(self, text=_('No node group selected')).grid(row=0, column=0, sticky='w')
            return

        self.bar = ttk.Frame(self, relief='raised')
        self.bar.pack(side='top', fill='both', pady=(0, 5))

        self.add_menu = tk.Menu(self.bar)
        for name in plugin.node_registry.keys():
            self.add_menu.add_command(label=name, command=partial(self.add_node, name))
        self.add_btn = ttk.Menubutton(self.bar, text=_('Add Node'), menu=self.add_menu)
        self.add_btn.grid(row=0, column=0, sticky='w')

        self.canvas = tk.Canvas(self, background='white')
        self.canvas.pack(side='left', fill='both', expand=True)

        self.draw_grid()

        self.dragging = False

        self.canvas.bind('<ButtonPress-1>', self.on_press)
        self.canvas.bind('<ButtonRelease-1>', self.on_release)
        self.canvas.bind('<B1-Motion>', self.on_move)
        self.canvas.bind('<MouseWheel>', self.on_scroll)

        # TODO: position and scale the view so all the nodes fit inside it
        row = self.conn.execute('''
            select min(x), min(y), max(x), max(y) from node where nodegroup_id = ?
        ''', (self.id, )).fetchone()
        assert row is not None
        xmin, ymin, xmax, ymax = row
        #self.canvas.scan_dragto(xmin - 10, ymin - 10)

        for id, name in self.conn.execute('select id, type from node where nodegroup_id = ?', (self.id, )):
            self.node_widgets[id] = NodeWidget(self.canvas, self.conn, self.listener, id)

        def setup_delayed() -> None:
            self.setup_sockets()
            self.setup_links()

        self.after(100, setup_delayed)

        self.group_io = NodeGroupIOSettings(self, self.conn, self.listener, self.id)
        self.group_io.pack(side='left', fill='both')

        self.listener.unregister(self.on_update)
        self.listener.unregister(self.on_delete)
        self.listener.register(self.on_update, apsw.SQLITE_UPDATE, 'main', 'nodegroup', self.id)
        self.listener.register(self.on_delete, apsw.SQLITE_DELETE, 'main', 'nodegroup', self.id)

    def setup_sockets(self) -> None:
        self.canvas.delete('socket')
        for node in self.node_widgets.values():
            size = 10
            x = node.winfo_rootx() - self.canvas.winfo_rootx() - 1 - size/2
            for id, label in node.input_labels.items():
                y = label.winfo_rooty() - self.canvas.winfo_rooty() + label.winfo_height() / 2 - size/2
                tags = (node.tag, 'socket', 'input_socket', f'socket-{id}')
                self.canvas.create_oval(x, y, x + size, y + size, fill='gray25', outline='gray25', tags=tags)

            x += node.winfo_width() + 2
            for id, label in node.output_labels.items():
                y = label.winfo_rooty() - self.canvas.winfo_rooty() + label.winfo_height() / 2 - size/2
                tags = (node.tag, 'socket', 'output_socket', f'socket-{id}')
                self.canvas.create_oval(x, y, x + size, y + size, fill='gray25', outline='gray25', tags=tags)

    def setup_links(self) -> None:
        for rowid, from_id, to_id in self.conn.execute('''
            select link.rowid, link.from_id, link.to_id
            from link
                join node_output o on link.from_id = o.id
                join node_input i on link.to_id = i.id
                join node on i.node_id = node.id
            where node.nodegroup_id = ?
        ''', (self.id, )).fetchall():
            assert isinstance(rowid, int) and isinstance(from_id, int) and isinstance(to_id, int)
            x1, y1, x2, y2 = self.canvas.coords(f'output_socket && socket-{from_id}')
            from_x = (x2 + x1) / 2
            from_y = (y2 + y1) / 2

            x1, y1, x2, y2 = self.canvas.coords(f'input_socket && socket-{to_id}')
            to_x = (x2 + x1) / 2
            to_y = (y2 + y1) / 2

            tags = ('link', f'link_from-{from_id}', f'link_to-{to_id}', f'link-{rowid}')
            self.canvas.create_line(from_x, from_y, to_x, to_y, fill='black', tags=tags)

    def add_node(self, name: str) -> None:
        assert self.id is not None
        with undogroup(self.conn):
            node = plugin.node_registry[name].create(self.conn, self.id)
        self.node_widgets[node.id] = NodeWidget(self.canvas, self.conn, self.listener, node.id)
        self.update()
        self.setup_sockets()

    @override
    def destroy(self) -> None:
        self.listener.unregister(self.on_update)
        self.listener.unregister(self.on_delete)
        self.listener.unregister(self.on_link_change)
        self.listener.unregister(self.on_selected_update)
        super().destroy()

    def on_update(self, _type: int, _db: str, _table: str, _rowid: int) -> None:
        pass

    def on_delete(self, _type: int, _db: str, _table: str, _rowid: int) -> None:
        self.setup()

    def on_selected_update(self, _type: int, _db: str, _table: str, _rowid: int) -> None:
        self.setup()

    def on_link_change(self, type: int, _db: str, _table: str, rowid: int) -> None:
        if type == apsw.SQLITE_DELETE:
            self.canvas.delete(f'link-{rowid}')
            return

        row = self.conn.execute('''
            select node.nodegroup_id, o.node_id, i.node_id, link.from_id, link.to_id
            from link
                join node_output o on link.from_id = o.id
                join node_input i on link.to_id = i.id
                join node on i.node_id = node.id
            where link.rowid = ?
        ''', (rowid, )).fetchone()
        assert row is not None
        nodegroup_id, from_node_id, to_node_id, from_id, to_id = row
        if nodegroup_id != self.id:
            return

        if type == apsw.SQLITE_UPDATE:
            self.canvas.delete(f'link-{rowid}')

        x1, y1, x2, y2 = self.canvas.coords(f'output_socket && socket-{from_id}')
        from_x = (x2 + x1) / 2
        from_y = (y2 + y1) / 2

        x1, y1, x2, y2 = self.canvas.coords(f'input_socket && socket-{to_id}')
        to_x = (x2 + x1) / 2
        to_y = (y2 + y1) / 2

        tags = ('link', f'link_from-{from_id}', f'link_to-{to_id}', f'link-{rowid}')
        self.canvas.create_line(from_x, from_y, to_x, to_y, fill='black', tags=tags)

    def draw_grid(self) -> None:
        for x in range(-500, 1500, 20):
            for y in range(-500, 1500, 20):
                major = x % 100 == 0 and y % 100 == 0
                sz = 5 if major else 3
                tags = ('grid', 'grid_major' if major else 'grid_minor')
                self.canvas.create_oval(x, y, x + sz, y + sz, fill='gray65', outline='', tags=tags)

    def on_press(self, event: tk.Event) -> None:
        items = self.canvas.find_overlapping(event.x - 5, event.y - 5, event.x + 5, event.y + 5)
        if item := next((item for item in items if 'socket' in self.canvas.gettags(item)), None):
            tags = self.canvas.gettags(item)
            id = int(next(tag for tag in tags if tag.startswith('socket-')).split('-')[1])
            if 'input_socket' in tags:
                link_items = self.canvas.find_withtag(f'link_to-{id}')
                if len(link_items) > 0:
                    link_item = link_items[0]
                    tags = self.canvas.gettags(link_item)
                    from_tag = next(tag for tag in tags if tag.startswith('link_from-'))
                    to_tag = next(tag for tag in tags if tag.startswith('link_to-'))
                    from_id = int(from_tag.split('-')[1])
                    to_id = int(to_tag.split('-')[1])
                    self.canvas.itemconfigure(link_item, tags=('link', 'wip_link', from_tag))
                    with undogroup(self.conn):
                        self.conn.execute('''
                            delete from link where from_id = ? and to_id = ?
                        ''', (from_id, to_id))
            elif 'output_socket' in tags:
                x1, y1, x2, y2 = self.canvas.coords(item)
                x = (x1 + x2) / 2
                y = (y1 + y2) / 2
                tags = ('link', 'wip_link', f'link_from-{id}')
                self.canvas.create_line(x, y, event.x, event.y, fill='black', tags=tags)
            else:
                raise Exception('unreachable')
        else:
            self.dragging = True
            self.canvas.scan_mark(event.x, event.y)

    def on_release(self, event: tk.Event) -> None:
        self.dragging = False

        items = self.canvas.find_overlapping(event.x - 5, event.y - 5, event.x + 5, event.y + 5)
        if item := next((item for item in items if 'socket' in self.canvas.gettags(item)), None):
            tags = self.canvas.gettags(item)
            id = int(next(tag for tag in tags if tag.startswith('socket-')).split('-')[1])
            if 'input_socket' not in tags:
                return
            link_items = self.canvas.find_withtag('wip_link')
            if len(link_items) > 0:
                link_item = link_items[0]
                link_tags = self.canvas.gettags(link_item)
                from_tag = next(tag for tag in link_tags if tag.startswith('link_from-'))
                from_id = int(from_tag.split('-')[1])
                with undogroup(self.conn):
                    self.conn.execute('delete from link where to_id = ?', (id, ))
                    self.conn.execute('''
                        insert into link (from_id, to_id) values (?, ?)
                    ''', (from_id, id))
        self.canvas.delete('wip_link')

    def on_move(self, event: tk.Event) -> None:
        if self.dragging:
            self.canvas.scan_dragto(event.x, event.y, gain=1)
        link_items = self.canvas.find_withtag('wip_link')
        if len(link_items) > 0:
            link_item = link_items[0]
            x1, y1, x2, y2 = self.canvas.coords(link_item)
            self.canvas.coords(link_item, x1, y1, event.x, event.y)

    def on_scroll(self, event: tk.Event) -> None:
        self.canvas.scale('all', event.x, event.y, event.delta, event.delta)


class NodeGroupIOSettings(ttk.Notebook):
    def __init__(self, parent: tk.Misc, conn: apsw.Connection, listener: ChangeListener, id: int) -> None:
        super().__init__(parent)
        self.conn = conn
        self.listener = listener
        self.id = id

        self.inputs = ttk.Frame(self)
        self.add(self.inputs, text=_('Inputs'))

        self.inputs_add_btn = ttk.Button(self.inputs, text=_('Add'), command=self.add_input)
        self.inputs_add_btn.grid(row=0, column=0, sticky='ew')
        self.inputs_delete_btn = ttk.Button(self.inputs, text=_('Delete'), command=self.delete_input)
        self.inputs_delete_btn.grid(row=0, column=1, sticky='ew')

        self.inputs_tree = ttk.Treeview(self.inputs)
        self.inputs_tree.grid(row=1, column=0, columnspan=2, sticky='ew')
        self.inputs_tree.bind('<<TreeviewSelect>>', self.on_input_select)

        inputs_scroll = ttk.Scrollbar(self.inputs, command=self.inputs_tree.yview)
        self.inputs_tree['yscrollcommand'] = inputs_scroll.set
        inputs_scroll.grid(row=1, column=2, sticky='ns')

        for input_id, name in self.conn.execute('''
            select id, name from nodegroup_input where nodegroup_id = ? order by idx
        ''', (self.id, )).fetchall():
            assert isinstance(input_id, int) and isinstance(name, str)
            self.inputs_tree.insert('', 'end', input_id, text=name)

        self.input_settings = ttk.Frame(self.inputs)
        self.input_settings.grid(row=2, column=0, columnspan=3, sticky='nsew')
        self.selected_input: int | None = None
        self.setup_input()

        self.outputs = ttk.Frame(self)
        self.add(self.outputs, text=_('Outputs'))

        self.outputs_add_btn = ttk.Button(self.outputs, text=_('Add'), command=self.add_output)
        self.outputs_add_btn.grid(row=0, column=0, sticky='ew')
        self.outputs_delete_btn = ttk.Button(self.outputs, text=_('Delete'), command=self.delete_output)
        self.outputs_delete_btn.grid(row=0, column=1, sticky='ew')

        self.outputs_tree = ttk.Treeview(self.outputs)
        self.outputs_tree.grid(row=1, column=0, columnspan=2, sticky='ew')
        self.outputs_tree.bind('<<TreeviewSelect>>', self.on_output_select)

        outputs_scroll = ttk.Scrollbar(self.outputs, command=self.outputs_tree.yview)
        self.outputs_tree['yscrollcommand'] = outputs_scroll.set
        outputs_scroll.grid(row=1, column=2, sticky='ns')

        for output_id, name in self.conn.execute('''
            select id, name from nodegroup_output where nodegroup_id = ? order by idx
        ''', (self.id, )).fetchall():
            assert isinstance(output_id, int) and isinstance(name, str)
            self.outputs_tree.insert('', 'end', output_id, text=name)

        self.output_settings = ttk.Frame(self.outputs)
        self.output_settings.grid(row=2, column=0, columnspan=3, sticky='nsew')
        self.selected_output: int | None = None
        self.setup_output()

        self.listener.register(self.on_input_change, None, 'main', 'nodegroup_input', None)
        self.listener.register(self.on_output_change, None, 'main', 'nodegroup_output', None)

    @override
    def destroy(self) -> None:
        self.listener.unregister(self.on_input_change)
        self.listener.unregister(self.on_output_change)
        super().destroy()

    def on_input_change(self, type: int, _db: str, _table: str, rowid: int) -> None:
        if type == apsw.SQLITE_DELETE:
            self.inputs_tree.delete(rowid)
            return

        row = self.conn.execute('''
            select nodegroup_id, name, idx from nodegroup_input where id = ?
        ''', (rowid, )).fetchone()
        assert row is not None
        nodegroup_id, name, idx = row
        if nodegroup_id != self.id:
            return

        if type == apsw.SQLITE_UPDATE:
            self.inputs_tree.item(rowid, text=name)
            self.inputs_tree.move(rowid, '', idx)
        elif type == apsw.SQLITE_INSERT:
            self.inputs_tree.insert('', idx, rowid, text=name)

    def on_output_change(self, type: int, _db: str, _table: str, rowid: int) -> None:
        if type == apsw.SQLITE_DELETE:
            self.outputs_tree.delete(rowid)
            return

        row = self.conn.execute('''
            select nodegroup_id, name, idx from nodegroup_output where id = ?
        ''', (rowid, )).fetchone()
        assert row is not None
        nodegroup_id, name, idx = row
        if nodegroup_id != self.id:
            return

        if type == apsw.SQLITE_UPDATE:
            self.outputs_tree.item(rowid, text=name)
            self.outputs_tree.move(rowid, '', idx)
        elif type == apsw.SQLITE_INSERT:
            self.outputs_tree.insert('', idx, rowid, text=name)

    def on_input_select(self, _event: tk.Event) -> None:
        if selected := self.inputs_tree.selection():
            self.selected_input = int(selected[-1])
            self.setup_input()

    def on_output_select(self, _event: tk.Event) -> None:
        if selected := self.outputs_tree.selection():
            self.selected_output = int(selected[-1])
            self.setup_output()

    def setup_input(self) -> None:
        for child in self.input_settings.winfo_children():
            child.destroy()

        if self.selected_input is None:
            input_lbl = ttk.Label(self.input_settings, text=_('No input selected'))
            input_lbl.grid(row=0, column=0, sticky='w')
            return

        row = self.conn.execute('''
            select name, idx, default_value, control_type from nodegroup_input where id = ?
        ''', (self.selected_input, )).fetchone()
        assert row is not None
        name, idx, default_value, control_type = row

        row = self.conn.execute('''
            select max(idx) from nodegroup_input where nodegroup_id = ?
        ''', (self.id, )).fetchone()
        assert row is not None
        max_idx, = row

        # TODO: fix all of this

        up_btn = ttk.Button(self.input_settings, text=_('Move Up'), command=self.move_input_up, state='disabled' if idx == 0 else 'normal')
        up_btn.grid(row=0, column=0, sticky='ew')
        down_btn = ttk.Button(self.input_settings, text=_('Move Down'), command=self.move_input_down, state='disabled' if idx == max_idx else 'normal')
        down_btn.grid(row=0, column=1, sticky='ew')

        self.input_name_var = tk.StringVar(value=name)
        name_lbl = ttk.Label(self.input_settings, text=_('Name:'))
        name_lbl.grid(row=1, column=0, sticky='w')
        name_entry = ttk.Entry(self.input_settings, textvariable=self.input_name_var)
        name_entry.grid(row=1, column=1, sticky='ew')

        self.input_default_value_var = tk.StringVar(value=default_value)
        default_value_lbl = ttk.Label(self.input_settings, text=_('Default value:'))
        default_value_lbl.grid(row=2, column=0, sticky='w')
        default_value_entry = ttk.Entry(self.input_settings, textvariable=self.input_default_value_var)
        default_value_entry.grid(row=2, column=1, sticky='ew')

        self.input_control_type_var = tk.StringVar(value=control_type)
        control_type_lbl = ttk.Label(self.input_settings, text=_('Control type:'))
        control_type_lbl.grid(row=3, column=0, sticky='w')
        control_type_combobox = ttk.Combobox(self.input_settings, textvariable=self.input_control_type_var, values=['checkbox', 'spinbox', 'entry'])
        control_type_combobox.grid(row=3, column=1, sticky='ew')

    def move_input_up(self) -> None:
        if self.selected_input is None:
            return
        with undogroup(self.conn):
            self.conn.execute('''
                with
                    selected as (select * from nodegroup_input where id = :id),
                    prev as (
                        select * from nodegroup_input
                        where nodegroup_id = :gid and idx < (select idx from selected)
                        order by idx desc limit 1
                    )
                update nodegroup_input set idx =
                    case when (select id from prev) not null then
                        case id
                            when :id then (select idx from prev)
                            when (select id from prev) then (select idx from selected)
                            else idx
                        end
                    else
                        idx
                    end
            ''', {'gid': self.id, 'id': self.selected_input})

    def move_input_down(self) -> None:
        if self.selected_input is None:
            return
        with undogroup(self.conn):
            self.conn.execute('''
                with
                    last as (select max(idx) idx from nodegroup_input where nodegroup_id = :gid),
                    selected as (select * from nodegroup_input where id = :id),
                    next as (
                        select * from nodegroup_input
                        where nodegroup_id = :gid and idx > (select idx from selected)
                        order by idx asc limit 1
                    )
                update nodegroup_input set idx =
                    case when (select id from next) not null then
                        case id
                            when :id then (select idx from next)
                            when (select id from next) then (select idx from selected)
                            else idx
                        end
                    else
                        idx
                    end
            ''', {'gid': self.id, 'id': self.selected_input})

    def setup_output(self) -> None:
        for child in self.output_settings.winfo_children():
            child.destroy()

        if self.selected_output is None:
            output_lbl = ttk.Label(self.output_settings, text=_('No output selected'))
            output_lbl.grid(row=0, column=0, sticky='w')
            return

        row = self.conn.execute('''
            select name, idx from nodegroup_output where id = ?
        ''', (self.selected_output, )).fetchone()
        assert row is not None
        name, idx = row

        row = self.conn.execute('''
            select max(idx) from nodegroup_output where nodegroup_id = ?
        ''', (self.id, )).fetchone()
        assert row is not None
        max_idx, = row

        up_btn = ttk.Button(self.output_settings, text=_('Move Up'), command=self.move_output_up, state='disabled' if idx == 0 else 'normal')
        up_btn.grid(row=0, column=0, sticky='ew')
        down_btn = ttk.Button(self.output_settings, text=_('Move Down'), command=self.move_output_down, state='disabled' if idx == max_idx else 'normal')
        down_btn.grid(row=0, column=1, sticky='ew')

        self.output_name_var = tk.StringVar(value=name)
        name_lbl = ttk.Label(self.output_settings, text=_('Name:'))
        name_lbl.grid(row=1, column=0, sticky='w')
        name_entry = ttk.Entry(self.output_settings, textvariable=self.output_name_var)
        name_entry.grid(row=1, column=1, sticky='ew')

    def move_output_up(self) -> None:
        if self.selected_output is None:
            return
        with undogroup(self.conn):
            self.conn.execute('''
                with
                    selected as (select * from nodegroup_output where id = :id),
                    prev as (
                        select * from nodegroup_output
                        where nodegroup_id = :gid and idx < (select idx from selected)
                        order by idx desc limit 1
                    )
                update nodegroup_output set idx =
                    case when (select id from prev) not null then
                        case id
                            when :id then (select idx from prev)
                            when (select id from prev) then (select idx from selected)
                            else idx
                        end
                    else
                        idx
                    end
            ''', {'gid': self.id, 'id': self.selected_output})

    def move_output_down(self) -> None:
        if self.selected_output is None:
            return
        with undogroup(self.conn):
            self.conn.execute('''
                with
                    last as (select max(idx) idx from nodegroup_output where nodegroup_id = :gid),
                    selected as (select * from nodegroup_output where id = :id),
                    next as (
                        select * from nodegroup_output
                        where nodegroup_id = :gid and idx > (select idx from selected)
                        order by idx asc limit 1
                    )
                update nodegroup_output set idx =
                    case when (select id from next) not null then
                        case id
                            when :id then (select idx from next)
                            when (select id from next) then (select idx from selected)
                            else idx
                        end
                    else
                        idx
                    end
            ''', {'gid': self.id, 'id': self.selected_output})

    def add_input(self) -> None:
        with undogroup(self.conn):
            self.conn.execute('''
                with next_idx as (
                    select coalesce(max(idx), -1) + 1 as idx
                        from nodegroup_input
                        where nodegroup_id = :id
                )
                insert into nodegroup_input (nodegroup_id, name, idx) values (
                    :id,
                    'Input '||(select idx + 1 from next_idx),
                    (select idx from next_idx)
                )
            ''', {'id': self.id})
        logger.info(f'Created new node group input with id {self.conn.last_insert_rowid()}')

    def delete_input(self) -> None:
        if selected := self.inputs_tree.selection():
            with undogroup(self.conn):
                for id in selected:
                    self.conn.execute('delete from nodegroup_input where id = ?', (id, ))

    def add_output(self) -> None:
        with undogroup(self.conn):
            self.conn.execute('''
                with next_idx as (
                    select coalesce(max(idx), -1) + 1 as idx
                        from nodegroup_output
                        where nodegroup_id = :id
                )
                insert into nodegroup_output (nodegroup_id, name, idx) values (
                    :id,
                    'Output '||(select idx + 1 from next_idx),
                    (select idx from next_idx)
                )
            ''', {'id': self.id})
        logger.info(f'Created new node group input with id {self.conn.last_insert_rowid()}')

    def delete_output(self) -> None:
        if selected := self.outputs_tree.selection():
            with undogroup(self.conn):
                for id in selected:
                    self.conn.execute('delete from nodegroup_output where id = ?', (id, ))


class NodeWidget(ttk.Frame):
    def __init__(self, canvas: tk.Canvas, conn: apsw.Connection, listener: ChangeListener, id: int) -> None:
        super().__init__(canvas, padding=5, relief='raised')
        self.canvas = canvas
        self.conn = conn
        self.listener = listener
        self.id = id

        row = self.conn.execute('select type, x, y from node where id = ?', (self.id, )).fetchone()
        assert row is not None
        type, x, y = row

        self.dragging = False

        self.tag = f'node-{self.id}'
        self.canvas.create_window(x, y, anchor='nw', window=self, tags=(self.tag, ))

        self.name_lbl = ttk.Label(self, text=type, font=('normal', 10, 'bold'))
        self.name_lbl.grid(row=0, column=0, columnspan=2, sticky='w')
        row = 1

        self.bind('<ButtonPress-1>', self.on_press)
        self.bind('<ButtonRelease-1>', self.on_release)
        self.bind('<B1-Motion>', self.on_move)
        self.name_lbl.bind('<ButtonPress-1>', self.on_press)
        self.name_lbl.bind('<ButtonRelease-1>', self.on_release)
        self.name_lbl.bind('<B1-Motion>', self.on_move)

        self.output_labels: dict[int, tk.Misc] = {}
        for output_id, name, idx in self.conn.execute('''
            select id, name, idx from node_output where node_id = ? order by idx
        ''', (self.id, )).fetchall():
            assert isinstance(output_id, int) and isinstance(name, str) and isinstance(idx, int)

            label = ttk.Label(self, text=name)
            label.grid(row=row, column=0, columnspan=2, sticky='e', padx=(30, 0))
            self.output_labels[output_id] = label
            self.l = label

            self.listener.register(self.on_output_change, apsw.SQLITE_UPDATE, 'main', 'node_output', output_id)
            row += 1

        self.input_labels: dict[int, tk.Misc] = {}
        self.input_controls: list[tk.Misc | None] = []
        for input_id, name, control_type, control_info, value, has_link in self.conn.execute('''
            select
                id,
                name,
                control_type,
                control_info,
                value,
                exists (select 1 from link where to_id = node_input.id) as has_link
            from node_input
            where node_id = ?
            order by idx
        ''', (self.id, )).fetchall():
            assert isinstance(input_id, int) and isinstance(name, str)
            assert isinstance(control_info, str) or control_info is None
            info: dict[str, Any] = {} if control_info is None else json.loads(control_info)

            pady = (10, 0) if len(self.input_labels) == 0 else 0

            label = ttk.Label(self, text=name if has_link else _(f'{name}:'))
            label.grid(row=row, column=0, sticky='w', pady=pady)
            self.input_labels[input_id] = label

            self.listener.register(self.on_input_change, apsw.SQLITE_UPDATE, 'main', 'node_input', input_id)

            if has_link: # TODO: make this reactive
                row += 1
                continue

            var = tk.StringVar(value='' if value is None else str(value))
            var_type: Callable[[str], bool | int | float | str | None] = str

            control: tk.Widget | None = None
            if control_type == 'checkbox':
                var_type = bool
                control = ttk.Checkbutton(self, variable=var)
            elif control_type == 'spinbox':
                var_type = float
                control = ttk.Spinbox(self, from_=-2**31, to=2**31-1, increment=1, width=8, textvariable=var)
            elif control_type == 'entry':
                var_type = str
                control = ttk.Entry(self, textvariable=var)
            elif control_type == 'combobox':
                var_type = str
                control = ttk.Combobox(self, values=info['values'], textvariable=var)
            if control:
                control.grid(row=row, column=1, sticky='w', padx=(0, 40), pady=pady)
            self.input_controls.append(control)

            def on_change(input_id: int, var: tk.StringVar, var_type: Callable[[str], bool | int | float | str | None], _var: str, _index: str, _mode: str) -> None:
                with undogroup(self.conn):
                    self.conn.execute('''
                        update node_input set value = ? where id = ?
                    ''', (var_type(var.get()), input_id))

            var.trace_add('write', partial(on_change, input_id, var, var_type))

            row += 1

        self.listener.register(self.on_update, apsw.SQLITE_UPDATE, 'main', 'node', self.id)
        self.listener.register(self.on_input_change, None, 'main', 'node_input', None)
        self.listener.register(self.on_output_change, None, 'main', 'node_output', None)

    @override
    def destroy(self) -> None:
        super().destroy()
        self.listener.unregister(self.on_update)
        self.listener.unregister(self.on_input_change)
        self.listener.unregister(self.on_output_change)

    def on_input_change(self, type: int, _db: str, _table: str, rowid: int) -> None:
        row = self.conn.execute('''
            select node_id, name, idx, control_type from node_input where id = ?
        ''', (rowid, )).fetchone()
        assert row is not None
        node_id, name, idx, control_type = row
        if node_id != self.id:
            return

        if type == apsw.SQLITE_INSERT:
            pass # TODO: dynamic inputs
        elif type == apsw.SQLITE_UPDATE:
            self.input_labels[rowid]['text'] = name
            # TODO: handle dynamic control type
        elif type == apsw.SQLITE_DELETE:
            pass

    def on_output_change(self, type: int, _db: str, _table: str, rowid: int) -> None:
        row = self.conn.execute('''
            select node_id, name, idx from node_output where id = ?
        ''', (rowid, )).fetchone()
        assert row is not None
        node_id, name, idx = row
        if node_id != self.id:
            return

        if type == apsw.SQLITE_INSERT:
            pass # TODO: dynamic inputs
        elif type == apsw.SQLITE_UPDATE:
            self.output_labels[rowid]['text'] = name
            # TODO: handle dynamic control type
        elif type == apsw.SQLITE_DELETE:
            pass

    def on_update(self, _type: int, _db: str, _table: str, _rowid: int) -> None:
        row = self.conn.execute('select type, x, y from node where id = ?', (self.id, )).fetchone()
        assert row is not None
        type, x, y = row
        self.canvas.moveto(self.tag, x, y)

    def on_press(self, event: tk.Event) -> None:
        self.dragging = True

    def on_release(self, event: tk.Event) -> None:
        if not self.dragging:
            return
        self.dragging = False
        x, y = self.canvas.coords(self.tag)
        with undogroup(self.conn):
            self.conn.execute('update node set x = ?, y = ? where id = ?', (x, y, self.id))
        logger.info(f'Finished dragging node with id {self.id}')

    def on_move(self, event: tk.Event) -> None:
        if not self.dragging:
            return
        self.canvas.move(self.tag, event.x, event.y)

        for id in self.input_labels.keys():
            socket, = self.canvas.find_withtag(f'input_socket && socket-{id}')
            x1, y1, x2, y2 = self.canvas.coords(socket)
            x = (x2 + x1) / 2
            y = (y2 + y1) / 2
            for item in self.canvas.find_withtag(f'link_to-{id}'):
                x1, y1, x2, y2 = self.canvas.coords(item)
                self.canvas.coords(item, x1, y1, x, y)

        for id in self.output_labels.keys():
            socket, = self.canvas.find_withtag(f'output_socket && socket-{id}')
            x1, y1, x2, y2 = self.canvas.coords(socket)
            x = (x2 + x1) / 2
            y = (y2 + y1) / 2
            for item in self.canvas.find_withtag(f'link_from-{id}'):
                x1, y1, x2, y2 = self.canvas.coords(item)
                self.canvas.coords(item, x, y, x2, y2)
