# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import json
import logging
import tkinter as tk
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import override, final, Type, Self, Any, Callable
import apsw
import pandas as pd


logger = logging.getLogger(__name__)


class Node(ABC):
    name = ''

    @classmethod
    @abstractmethod
    def create(cls, conn: apsw.Connection, nodegroup_id: int) -> Self:
        pass

    def __init__(self, conn: apsw.Connection, id: int) -> None:
        self.conn = conn
        self.id = id

    @abstractmethod
    def run(self, inputs: list[Any]) -> list[Any]:
        pass

    @final
    def set_control_info(self, input_name: str, info: Any) -> None:
        with self.conn:
            self.conn.execute('update trigger_settings set enabled = 0')
            self.conn.execute('''
                update node_input set control_info = ? where node_id = ? and name = ?
            ''', (json.dumps(info), self.id, input_name))
            self.conn.execute('update trigger_settings set enabled = 1')

    @final
    def set_input_value(self, input_name: str, value: bool | int | float | str | None) -> None:
        with self.conn:
            self.conn.execute('update trigger_settings set enabled = 0')
            self.conn.execute('''
                update node_input set value = ? where node_id = ? and name = ?
            ''', (value, self.id, input_name))
            self.conn.execute('update trigger_settings set enabled = 1')


class BasicNode(Node):
    inputs: list[tuple[str, str | None, Any, Any]] = []
    outputs: list[str] = []

    @final
    @override
    @classmethod
    def create(cls, conn: apsw.Connection, nodegroup_id: int) -> Self:
        with conn:
            conn.execute('''
                insert into node (nodegroup_id, type) values (?, ?)
            ''', (nodegroup_id, cls.name))
            id = conn.last_insert_rowid()
            for i, (name, control_type, control_info, value) in enumerate(cls.inputs):
                control_info = json.dumps(control_info) if control_info is not None else None
                conn.execute('''
                    insert into node_input (node_id, name, idx, control_type, control_info, value)
                        values (?, ?, ?, ?, ?, ?)
                ''', (id, name, i, control_type, control_info, value))
            for i, name in enumerate(cls.outputs):
                conn.execute('''
                    insert into node_output (node_id, name, idx) values (?, ?, ?)
                ''', (id, name, i))
        return cls(conn, id)


node_registry: dict[str, Type[Node]] = {}


def register_node(cls: Type[Node]) -> Type[Node]:
    node_registry[cls.name] = cls
    return cls


type FileLoader = Callable[[apsw.Blob], pd.DataFrame]


loader_registry: dict[str, FileLoader] = {}


def register_loader(filetype: str) -> Callable[[FileLoader], FileLoader]:
    def f(func: FileLoader) -> FileLoader:
        loader_registry[filetype] = func
        return func
    return f
