# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import collections
import logging
from typing import Callable
import apsw


logger = logging.getLogger(__name__)


type ChangeCallback = Callable[[int, str, str, int], None]
type ChangePattern = tuple[int | None, str | None, str | None, int | None]


class ChangeListener:
    def __init__(self, conn: apsw.Connection) -> None:
        self._conn = conn
        self._listeners: dict[ChangePattern, set[ChangeCallback]] = collections.defaultdict(set)
        self._changes: list[tuple[int, str, str, int]] = []
        self._conn.set_update_hook(self._hook)
        self._conn.trace_v2(apsw.SQLITE_TRACE_PROFILE, self._handle_changes)

    def _hook(self, type: int, db: str, table: str, rowid: int) -> None:
        self._changes.append((type, db, table, rowid))

    def _handle_changes(self, info: dict) -> None:
        while len(self._changes) > 0:
            key = self._changes.pop()
            patterns = []
            for mask in range(2**4):
                pattern: ChangePattern = tuple(x if mask & (1 << i) else None for i, x in enumerate(key)) # type: ignore
                patterns.append(pattern)
            for pattern in patterns:
                for callback in self._listeners[pattern]:
                    callback(*key)

    def register(self, callback: ChangeCallback, type: int | None = None, db: str | None = None, table: str | None = None, rowid: int | None = None) -> None:
        self._listeners[(type, db, table, rowid)].add(callback)
        logger.debug(f'Registered change listener {callback} for type={type}, db={db}, table={table}, rowid={rowid}')

    def unregister(self, callback: ChangeCallback) -> None:
        for callbacks in self._listeners.values():
            callbacks.discard(callback)
        logger.debug(f'Unregistered change listener {callback}')
