# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import json
import logging
from typing import Any
import apsw
import pandas as pd
from biomeca.plugin import node_registry, loader_registry


logger = logging.getLogger(__name__)


def run_nodegroup(conn: apsw.Connection, id: int, inputs: list[Any]) -> list[Any]:
    outputs: dict[int, Any] = {}

    for output_id, idx in conn.execute('''
        select o.id, o.idx
        from node_output o
            join node on o.node_id = node.id
        where node.nodegroup_id = ? and node.type = 'Group Input'
    ''', (id, )).fetchall():
        assert isinstance(output_id, int) and isinstance(idx, int)
        outputs[output_id] = inputs[idx]

    while nodes := conn.execute('''
        select id, subgroup_id, type from node
        where nodegroup_id = :id and type != 'Group Input'
            and not exists (
                select 1
                from node_input
                    join link on link.to_id = node_input.id
                where node_input.node_id = node.id
                    and link.from_id not in (select value from json_each(:finished))
            )
            and (type = 'Group Output' or exists (
                select id from node_output where node_id = node.id
                except select value from json_each(:finished)
            ))
    ''', {'id': id, 'finished': json.dumps(list(outputs.keys()))}).fetchall():
        for node_id, subgroup_id, type in nodes:
            node_inputs = conn.execute('''
                select value, from_id
                from node_input
                    left join link on link.to_id = node_input.id
                where node_id = ?
                order by idx
            ''', (node_id, )).fetchall()
            assert isinstance(node_id, int)
            assert isinstance(subgroup_id, int) or subgroup_id is None
            assert isinstance(type, str) or type is None

            input_values = []
            for value, from_id in node_inputs:
                assert from_id is None or isinstance(from_id, int)
                input_values.append(value if from_id is None else outputs.get(from_id, value))
            if type == 'Group Output':
                return input_values
            if subgroup_id is not None:
                output_values = run_nodegroup(conn, subgroup_id, input_values)
            elif type is not None:
                output_values = node_registry[type](conn, node_id).run(input_values)
            else:
                assert False, 'unreachable'

            node_outputs = conn.execute('''
                select id from node_output where node_id = ? order by idx
            ''', (node_id, )).fetchall()

            for (output_id, ), value in zip(node_outputs, output_values):
                assert isinstance(output_id, int)
                outputs[output_id] = value
    return []


def run_pane(conn: apsw.Connection, id: int) -> Any:
    row = conn.execute('select nodegroup_id from pane where id = ?', (id, )).fetchone()
    if row is None or row[0] is None:
        return None
    nodegroup_id, = row

    inputs: list[Any] = []

    data = {}
    for file_id, name, type in conn.execute('select id, name, type from file').fetchall():
        assert isinstance(file_id, int) and isinstance(name, str) and isinstance(type, str)
        with conn.blob_open('main', 'file', 'content', file_id, writeable=False) as blob:
            if loader := loader_registry.get(type):
                data[name] = loader(blob)
            else:
                logger.warn(f'Unable to find a loader for filetype {type}')
    inputs.append(pd.concat(data, names=['File', 'Label', 'Dimension'], axis='columns') if data else None)

    for input_value, in conn.execute('''
        select p.value
        from pane_input p
            join nodegroup_input n on p.nodegroup_input_id = n.id
        where p.pane_id = ?
        order by n.idx
    ''', (id, )).fetchall():
        inputs.append(input_value)

    outputs = run_nodegroup(conn, nodegroup_id, inputs)
    return outputs[0] if len(outputs) > 0 else None
