# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only


import contextlib
import logging
from typing import Generator
import apsw


logger = logging.getLogger(__name__)


def setup_undo(conn: apsw.Connection, tables: list[str] | None = None) -> None:
    if not tables:
        tables = []
        for name, in conn.execute("""
            select name from main.sqlite_schema where type = 'table'
        """).fetchall():
            assert isinstance(name, str)
            if not name.startswith('sqlite_'):
                tables.append(name)

    logger.debug(f'Setting up undo/redo for the following tables: {tables}')
    with conn:
        conn.execute('''
            create temp table undostack(
                id integer primary key,
                current integer unique check(current is null or current = 1),
                unfinished integer unique,
                undo_stmt text not null default '',
                redo_stmt text not null default '',
                check(unfinished is null or (unfinished = 1 and current = 1))
            );
            create index undostack_current on undostack(current);

            insert into undostack default values;
            insert into undostack (current, unfinished) values (1, 1);

            create temp table trigger_settings(enabled integer not null);
            insert into trigger_settings (enabled) values (1);
        ''')
        for table in tables:
            columns = ['rowid']
            for col, in conn.execute('select name from pragma_table_info(?)', (table, )).fetchall():
                assert isinstance(col, str)
                columns.append(col)
            # Note: apparently you can specify both the rowid and the integer primary key column
            conn.execute(f'''
create temp trigger {table}_insert_undo after insert on main.{table}
when (select enabled from trigger_settings) = 1 begin
    delete from undostack where id > (select id from undostack where current = 1);
    update undostack set current = null where unfinished is null;
    insert or ignore into undostack (current, unfinished) values (1, 1);
    update undostack set
        undo_stmt = 'delete from {table} where rowid = '||quote(new.rowid)||';'||undo_stmt,
        redo_stmt = redo_stmt||'insert into {table} ({', '.join(columns)}) values ('
            {"||', '\n".join(f'||quote(new.{col})' for col in columns)}
        ||');'
    where unfinished = 1;
end;

create temp trigger {table}_update_undo after update on main.{table}
when (select enabled from trigger_settings) = 1 begin
    delete from undostack where id > (select id from undostack where current = 1);
    update undostack set current = null where unfinished is null;
    insert or ignore into undostack (current, unfinished) values (1, 1);
    update undostack set
        undo_stmt = 'update {table} set '
            {"||', '\n".join(f"||'{col} = '||quote(old.{col})" for col in columns)}
        ||' where rowid = '||quote(new.rowid)||';'||undo_stmt,
        redo_stmt = redo_stmt||'update {table} set '
            {"||', '\n".join(f"||'{col} = '||quote(new.{col})" for col in columns)}
        ||' where rowid = '||quote(old.rowid)||';'
    where unfinished = 1;
end;

create temp trigger {table}_delete_undo before delete on main.{table}
when (select enabled from trigger_settings) = 1 begin
    delete from undostack where id > (select id from undostack where current = 1);
    update undostack set current = null where unfinished is null;
    insert or ignore into undostack (current, unfinished) values (1, 1);
    update undostack set
        undo_stmt = 'insert into {table} ({', '.join(columns)}) values ('
            {"||', '\n".join(f'||quote(old.{col})' for col in columns)}
        ||');'||undo_stmt,
        redo_stmt = redo_stmt||'delete from {table} where rowid = '||quote(old.rowid)||';'
    where unfinished = 1;
end;''')
    logger.debug('Finished setting up undo/redo')


def undo(conn: apsw.Connection) -> None:
    logger.debug('Undoing')
    row = conn.execute('select undo_stmt from undostack where current = 1').fetchone()
    assert row is not None
    undo_stmt, = row
    if not undo_stmt:
        logger.debug('Nothing to undo')
    with conn:
        if result := conn.execute('''
            select id from undostack
            where id < (select id from undostack where current = 1)
            order by id desc
        ''').fetchone():
            prev_id, = result
            conn.execute('update undostack set current = null where current = 1')
            conn.execute('update undostack set current = 1 where id = ?', (prev_id, ))
        conn.execute('update trigger_settings set enabled = 0')
        conn.execute(undo_stmt)
        conn.execute('update trigger_settings set enabled = 1')


def redo(conn: apsw.Connection) -> None:
    logger.debug('Redoing')
    if result := conn.execute('''
        select id, redo_stmt from undostack
        where id > (select id from undostack where current = 1)
        order by id
    ''').fetchone():
        next_id, redo_stmt = result
        with conn:
            conn.execute('update undostack set current = null where current = 1')
            conn.execute('update undostack set current = 1 where id = ?', (next_id, ))
            conn.execute('update trigger_settings set enabled = 0')
            conn.execute(redo_stmt)
            conn.execute('update trigger_settings set enabled = 1')
    else:
        logger.debug('Nothing to redo')


def finish_undogroup(conn: apsw.Connection) -> None:
    conn.execute('update undostack set unfinished = null where unfinished = 1')
    logger.debug('Finishing undogroup')


@contextlib.contextmanager
def undogroup(conn: apsw.Connection) -> Generator[None, None, None]:
    with conn:
        logger.debug('Entering undogroup')
        yield
        finish_undogroup(conn)
