# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import logging
import shutil
from gettext import gettext as _
from pathlib import Path
from tkinter import messagebox
from typing import IO
import apsw
from biomeca import undo


logger = logging.getLogger(__name__)


def format_size(size: int) -> str:
    if size < 1024**1:
        return _(f'{size/1024**0:.1f} B')
    elif size < 1024**2:
        return _(f'{size/1024**1:.1f} KiB')
    elif size < 1024**3:
        return _(f'{size/1024**2:.1f} MiB')
    elif size < 1024**4:
        return _(f'{size/1024**3:.1f} GiB')
    elif size < 1024**5:
        return _(f'{size/1024**4:.1f} TiB')
    elif size < 1024**6:
        return _(f'{size/1024**5:.1f} PiB')
    elif size < 1024**7:
        return _(f'{size/1024**6:.1f} EiB')
    elif size < 1024**8:
        return _(f'{size/1024**7:.1f} ZiB')
    elif size < 1024**9:
        return _(f'{size/1024**8:.1f} YiB')
    elif size < 1024**10:
        return _(f'{size/1024**9:.1f} RiB')
    elif size < 1024**11:
        return _(f'{size/1024**10:.1f} QiB')
    else:
        return _(f'Congratulations! You uploaded a file whose size is greater than the diameter of the observable universe in millimeters.')


def open_db(path: str) -> apsw.Connection:
    conn = apsw.Connection(path)

    with open(Path(__file__).parent / 'schema.sql') as file:
        schema = file.read()

    with conn:
        conn.config(apsw.SQLITE_DBCONFIG_DQS_DML, 0)
        conn.config(apsw.SQLITE_DBCONFIG_DQS_DDL, 0)
        conn.pragma('application_id', 0xb103ec5)
        version = conn.pragma('user_version')
        supported = 1
        if version == 0:
            conn.pragma('user_version', supported)
        elif version > supported:
            logger.error(f'Attempted to open file with unsupported version {version}')
            # TODO: find a way to have this function be independent of GUI stuff
            messagebox.showerror(_('Unsupported version'), _(f'This file\'s version is {version}, but only versions up to {supported} are supported. Please update the application to open this file.'))
        conn.pragma('foreign_keys', 'on')
        conn.pragma('trusted_schema', 'off')
        conn.pragma('recursive_triggers', 'off')
        conn.pragma('cell_size_check', 'on')
        conn.pragma('integrity_check')
        conn.pragma('optimize', 0x10002)
        conn.execute(schema)

    tables = [
        'file',
        'nodegroup',
        'nodegroup_input',
        'nodegroup_output',
        'node',
        'node_input',
        'node_output',
        'link',
        'pane',
        'pane_input',
    ]
    undo.setup_undo(conn, tables)

    logger.info(f'Opened project file at path {path!r}')
    return conn


def add_file(conn: apsw.Connection, path: Path) -> int:
    with conn:
        # If we use the incremental blob IO API, the change won't go into the undostack
        # TODO: do incremental io then 'update file set content = content where id = ?' to fix this
        with open(path, 'rb') as file:
            content = file.read()
        type = path.suffix.removeprefix('.')
        row = conn.execute('''
            insert into file (name, type, content) values (?, ?, ?) returning id
        ''', (path.stem, type, content)).fetchone()
        assert row is not None
        id, = row
        assert isinstance(id, int)
    logger.info(f'Added file at path {path!r} to the project')
    return id


def export_file(conn: apsw.Connection, id: int, file: IO) -> None:
    with conn.blobopen('main', 'file', 'content', id, False) as blob:
        shutil.copyfileobj(blob, file)
    logger.info(f'Exported file with rowid {id} to path {file.name!r}')
