# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import logging
import tkinter as tk
from gettext import gettext as _
from pathlib import Path
from tkinter import ttk, filedialog, messagebox, simpledialog
from typing import override
import apsw
import pandas as pd
from matplotlib.backend_bases import key_press_handler
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
from biomeca import undo, util, plugin, runtime
from biomeca.reactivity import ChangeListener
from biomeca.undo import undogroup
from biomeca.nodes import NodeEditor


logger = logging.getLogger(__name__)


class MainWindow:
    def __init__(self, root: tk.Tk = tk.Tk(), path: str = ':memory:') -> None:
        super().__init__()
        self.root = root

        self.conn = util.open_db(':memory:')
        self.listener = ChangeListener(self.conn)

        self.root.option_add('*tearOff', False)
        ttk.Style().theme_use('clam')

        menubar = tk.Menu(self.root)
        self.root['menu'] = menubar

        menu_file = tk.Menu(menubar)
        menubar.add_cascade(menu=menu_file, label=_('File'))
        menu_file.add_command(label=_('New Project'), accelerator=_('Ctrl+N'), command=self.new_project)
        menu_file.add_command(label=_('Open Project'), accelerator=_('Ctrl+O'), command=self.open_project)
        menu_file.add_command(label=_('Save'), accelerator=_('Ctrl+S'), command=self.save)
        menu_file.add_command(label=_('Save As'), accelerator=_('Ctrl+Shift+S'), command=self.save_as)
        menu_file.add_separator()
        menu_file.add_command(label=_('Data Sources...'), command=self.manage_files)
        menu_file.add_separator()
        menu_file.add_command(label=_('Quit'), accelerator=_('Ctrl+Q'), command=self.quit)
        self.root.bind_all('<Control-n>', lambda _: self.new_project())
        self.root.bind_all('<Control-o>', lambda _: self.open_project())
        self.root.bind_all('<Control-s>', lambda _: self.save())
        self.root.bind_all('<Control-Shift-s>', lambda _: self.save_as())
        self.root.bind_all('<Control-q>', lambda _: self.quit())

        menu_edit = tk.Menu(menubar)
        menubar.add_cascade(menu=menu_edit, label=_('Edit'))
        menu_edit.add_command(label=_('Undo'), accelerator=_('Ctrl+Z'), command=self.undo)
        menu_edit.add_command(label=_('Redo'), accelerator=_('Ctrl+Shift+Z'), command=self.redo)
        menu_edit.add_separator()
        menu_edit.add_command(label=_('Settings'), accelerator=_('Ctrl+,'), command=self.settings)
        self.root.bind_all('<Control-z>', lambda _: self.undo())
        self.root.bind_all('<Control-Shift-z>', lambda _: self.undo())
        self.root.bind_all('<Control-comma>', lambda _: self.settings())

        self.project_editor = ProjectEditor(self.root, self.conn, self.listener)
        self.setup_editor(path)

    def quit(self) -> None:
        self.conn.close()
        self.root.quit()

    def setup_editor(self, path: str = ':memory:') -> None:
        self.project_editor.destroy()
        self.conn.close()
        self.conn = util.open_db(path)
        self.listener = ChangeListener(self.conn)
        self.project_editor = ProjectEditor(self.root, self.conn, self.listener)
        self.project_editor.pack(side='left', fill='both', expand=1, padx=5, pady=5)

        row = self.conn.execute('''
            select file from pragma_database_list where name = 'main'
        ''').fetchone()
        assert row is not None
        path, = row
        self.root.wm_title(path or _('Unsaved'))

    def new_project(self) -> None:
        self.setup_editor()

    def open_project(self, path: str | None = None) -> None:
        if not path:
            path = filedialog.askopenfilename(parent=self.root, title=_('Open Project'), filetypes=[('Project Files', '*.sqlite')])
        if not path:
            return
        try:
            self.setup_editor(path)
        except Exception as e:
            logger.exception(f'Error opening project at path {path!r}')
            messagebox.showerror(_('Error opening project'), _(f'Error: unable to open project: {str(e)}'))
            return

    def save(self) -> None:
        row = self.conn.execute('''
            select file from pragma_database_list where name = 'main'
        ''').fetchone()
        assert row is not None
        path, = row
        if path:
            return
        if path := self.save_as():
            self.open_project(path)

    def save_as(self) -> str | None:
        if path := filedialog.asksaveasfilename(parent=self.root, title=_('Save As'), filetypes=[('Project Files', '*.sqlite')]):
            try:
                dest = apsw.Connection(path)
                with dest.backup('main', self.conn, 'main') as backup:
                    while not backup.done:
                        backup.step(256)
            except Exception as e:
                logger.error(f'Error saving project to path {path!r}: {e!r}')
                messagebox.showerror(_('Error saving project'), _(f'Error: unable to save project: {str(e)}'))
                return None
            logger.info(f'Saved project to path {path!r}')
            return path
        return None

    def manage_files(self) -> None:
        logger.info('Opened files dialog')
        FilesDialog(self.root, self.conn, self.listener)

    def undo(self) -> None:
        logger.info('Undo')
        undo.undo(self.conn)

    def redo(self) -> None:
        logger.info('Redo')
        undo.redo(self.conn)

    def settings(self) -> None:
        logger.info('Opened settings dialog')
        SettingsDialog(self.root)


class SettingsDialog(simpledialog.Dialog):
    def __init__(self, parent: tk.Misc) -> None:
        self.theme = tk.StringVar(value=ttk.Style().theme_use())
        self.theme.trace_add('write', self.theme_changed)
        super().__init__(parent, title=_('Settings'))

    @override
    def body(self, parent: tk.Misc) -> None:
        notebook = ttk.Notebook(parent)
        notebook.grid(row=0, column=0, sticky='nsew')

        appearance = ttk.Frame(notebook, padding=10)
        notebook.add(appearance, text=_('Appearance'))
        ttk.Label(appearance, text=_('Theme:')).grid(row=0, column=0, padx=(0, 20), sticky='w')
        themes = ttk.Style().theme_names()
        theme_chooser = ttk.Combobox(appearance, textvariable=self.theme, values=themes)
        theme_chooser.grid(row=0, column=1, sticky='e')

    @override
    def buttonbox(self) -> None:
        ok = ttk.Button(self, text=_('Ok'), command=self.ok)
        ok.pack(side='bottom', anchor='e', padx=5, pady=5)

    def theme_changed(self, _var: str, _index: str, mode: str) -> None:
        theme = self.theme.get()
        logger.info(f'Changed theme to {theme!r}')
        ttk.Style().theme_use(theme)


class FilesDialog(simpledialog.Dialog):

    FILETYPES = [('C3D files', '*.c3d'), ('CSV files', '*.csv')]

    def __init__(self, parent: tk.Misc, conn: apsw.Connection, listener: ChangeListener) -> None:
        self.conn = conn
        self.listener = listener
        super().__init__(parent, title=_('Data Sources'))

    @override
    def body(self, parent: tk.Misc) -> None:
        toolbar = ttk.Frame(parent, relief='raised')
        toolbar.grid(row=0, column=0, columnspan=2, sticky='ew')
        tool_btns = [
            (_('Add'), self.add_files),
            (_('Rename'), self.rename),
            (_('Export'), self.export),
            (_('Delete'), self.delete),
        ]
        ttk.Style().configure('flat.TButton', relief='flat')
        for i, (text, command) in enumerate(tool_btns):
            btn = ttk.Button(toolbar, text=text, command=command, style='flat.TButton')
            btn.grid(row=0, column=i, pady=2, sticky='ew')

        self.files = ttk.Treeview(parent, show='headings')
        scroll = ttk.Scrollbar(parent, command=self.files.yview)
        self.files['yscrollcommand'] = scroll.set
        self.files.grid(row=1, column=0, sticky='nsew')
        scroll.grid(row=1, column=1, sticky='ns')

        self.files.bind('<Double-Button-1>', self.double_click_file)

        self.files['columns'] = ('File Name', 'Size', 'Type')
        self.files.column('File Name', width=150)
        self.files.heading('File Name', text=_('File Name'))
        self.files.column('Size', width=100)
        self.files.heading('Size', text=_('Size'))
        self.files.column('Type', width=100)
        self.files.heading('Type', text=_('Type'))

        for id, in self.conn.execute('select id from file').fetchall():
            assert isinstance(id, int)
            self.display_file(id)

        self.listener.register(self.on_insert, apsw.SQLITE_INSERT, 'main', 'file')
        self.listener.register(self.on_update, apsw.SQLITE_UPDATE, 'main', 'file')
        self.listener.register(self.on_delete, apsw.SQLITE_DELETE, 'main', 'file')

    @override
    def destroy(self) -> None:
        self.listener.unregister(self.on_insert)
        self.listener.unregister(self.on_update)
        self.listener.unregister(self.on_delete)
        super().destroy()

    @override
    def buttonbox(self) -> None:
        self.done_btn = ttk.Button(self, text=_('Done'), command=self.ok)
        self.done_btn.pack(side='right', padx=20, pady=10)

    def on_insert(self, _type: int, _db: str, _table: str, rowid: int) -> None:
        self.display_file(rowid)

    def on_update(self, _type: int, _db: str, _table: str, rowid: int) -> None:
        self.update_file(rowid)

    def on_delete(self, _type: int, _db: str, _table: str, rowid: int) -> None:
        self.files.delete(rowid)

    def display_file(self, id: int) -> None:
        self.files.insert('', 'end', id)
        self.update_file(id)

    def update_file(self, id: int) -> None:
        row = self.conn.execute('''
            select name, type, length(content) from file where id = ?
        ''', (id, )).fetchone()
        assert row is not None
        name, type, size = row
        sizestr = util.format_size(size)
        self.files.item(id, values=(name, sizestr, type))

    def double_click_file(self, event: tk.Event) -> None:
        if clicked_id := self.files.identify_row(event.y):
            self.files.selection_set(clicked_id)
            self.rename()

    def add_files(self) -> None:
        if paths := filedialog.askopenfilenames(filetypes=self.FILETYPES):
            with undogroup(self.conn):
                for path in map(Path, paths):
                    util.add_file(self.conn, path)

    def rename(self) -> None:
        selected = self.files.selection()
        if len(selected) == 0:
            messagebox.showwarning(_('No files selected'), _('You must select at least one file.'))
            return
        for id in selected:
            oldname = self.files.set(id, 'File Name')
            if name := simpledialog.askstring(_('Rename File'), _(f'New name for file {oldname}:')):
                with undogroup(self.conn):
                    self.conn.execute('update file set name = ? where id = ?', (name, id))
                logger.info(f'Renamed file {oldname!r} to {name!r}')

    def export(self) -> None:
        selected = self.files.selection()
        if len(selected) == 0:
            messagebox.showwarning(_('No files selected'), _('You must select at least one file.'))
            return
        for id in selected:
            row = self.conn.execute('select name, type from file where id = ?', (id, )).fetchone()
            assert row is not None
            name, type = row
            ftypes = [(name, glob) for name, glob in self.FILETYPES if glob.endswith('.' + type)]
            if fh := filedialog.asksaveasfile('wb', title=_(f'Export File "{name}"'), filetypes=ftypes, confirmoverwrite=True, initialfile=f'{name}.{type}'):
                util.export_file(self.conn, int(id), fh)

    def delete(self) -> None:
        selected = self.files.selection()
        if len(selected) == 0:
            messagebox.showwarning(_('No files selected'), _('You must select at least one file.'))
            return
        if not messagebox.askyesno(_('Confirm file deletion'), _('Are you sure you want to delete all of the selected files?')):
            return
        with undogroup(self.conn):
            for id in selected:
                self.conn.execute('delete from file where id = ?', (int(id), ))
                logger.info(f'Deleted file with id {id}')


class ProjectEditor(ttk.Frame):
    def __init__(self, parent: tk.Misc, conn: apsw.Connection, listener: ChangeListener) -> None:
        super().__init__(parent)
        self.conn = conn
        self.listener = listener

        self.leftwin = ttk.PanedWindow(self, orient='vertical')
        self.leftwin.pack(side='left', fill='both', expand=True, padx=(0, 5))

        self.panes_view = PanesViewport(self.leftwin, self.conn, self.listener)
        self.leftwin.add(self.panes_view, weight=3)
        self.node_editor = NodeEditor(self.leftwin, self.conn, self.listener)
        self.leftwin.add(self.node_editor, weight=2)

        self.rightwin = ttk.PanedWindow(self, orient='vertical')
        self.rightwin.pack(side='left', fill='both')

        self.panes_list = PanesList(self.rightwin, self.conn, self.listener)
        self.rightwin.add(self.panes_list, weight=1)
        self.pane_settings = PaneSettings(self.rightwin, self.conn, self.listener)
        self.rightwin.add(self.pane_settings, weight=2)


class PanesViewport(ttk.Frame):
    def __init__(self, parent: tk.Misc, conn: apsw.Connection, listener: ChangeListener) -> None:
        super().__init__(parent, relief='sunken')
        self.conn = conn
        self.listener = listener
        self.panes: dict[int, PaneViewer] = {}
        for id, in self.conn.execute('select id from pane').fetchall():
            assert isinstance(id, int)
            self.panes[id] = PaneViewer(self, self.conn, self.listener, id)

        self.listener.register(self.on_insert, apsw.SQLITE_INSERT, 'main', 'pane')
        self.listener.register(self.on_delete, apsw.SQLITE_DELETE, 'main', 'pane')

    @override
    def destroy(self) -> None:
        self.listener.unregister(self.on_insert)
        self.listener.unregister(self.on_delete)
        super().destroy()

    def on_insert(self, _type: int, _db: str, _table: str, rowid: int) -> None:
        self.panes[rowid] = PaneViewer(self, self.conn, self.listener, rowid)

    def on_delete(self, _type: int, _db: str, _table: str, rowid: int) -> None:
        self.panes.pop(rowid).destroy()


class PanesList(ttk.Frame):
    def __init__(self, parent: tk.Misc, conn: apsw.Connection, listener: ChangeListener) -> None:
        super().__init__(parent, relief='sunken', padding=5)
        self.conn = conn
        self.listener = listener

        toolbar = ttk.Frame(self)
        toolbar.grid(row=0, column=0, columnspan=2, sticky='ew', pady=(0, 5))

        add_btn = ttk.Button(toolbar, text=_('Add'), command=self.add_pane)
        add_btn.grid(row=0, column=0, padx=(0, 5))

        delete_btn = ttk.Button(toolbar, text=_('Delete'), command=self.delete_selected)
        delete_btn.grid(row=0, column=1)

        self.panes = ttk.Treeview(self)
        self.panes.grid(row=1, column=0, sticky='nsew')
        self.panes.bind('<<TreeviewSelect>>', self.on_select)

        scroll = ttk.Scrollbar(self, command=self.panes.yview)
        self.panes['yscrollcommand'] = scroll.set
        scroll.grid(row=1, column=1, sticky='ns')

        for id, title in self.conn.execute('select id, title from pane').fetchall():
            assert isinstance(id, int)
            assert isinstance(title, str)
            self.panes.insert('', 'end', id, text=title)

        row = self.conn.execute('select pane_id from selected_pane').fetchone()
        assert row is not None
        id, = row
        if id is not None:
            assert isinstance(id, int)
            if self.panes.exists(id):
                self.panes.selection_set(id)

        self.listener.register(self.on_insert, apsw.SQLITE_INSERT, 'main', 'pane')
        self.listener.register(self.on_update, apsw.SQLITE_UPDATE, 'main', 'pane')
        self.listener.register(self.on_delete, apsw.SQLITE_DELETE, 'main', 'pane')

    @override
    def destroy(self) -> None:
        self.listener.unregister(self.on_insert)
        self.listener.unregister(self.on_update)
        self.listener.unregister(self.on_delete)
        super().destroy()

    def on_insert(self, _type: int, _db: str, _table: str, rowid: int) -> None:
        row = self.conn.execute('select title from pane where id = ?', (rowid, )).fetchone()
        assert row is not None
        title, = row
        assert isinstance(title, str)
        self.panes.insert('', 'end', rowid, text=title)

    def on_update(self, _type: int, _db: str, _table: str, rowid: int) -> None:
        row = self.conn.execute('select title from pane where id = ?', (rowid, )).fetchone()
        assert row is not None
        title, = row
        assert isinstance(title, str)
        self.panes.item(rowid, text=title)

    def on_delete(self, _type: int, _db: str, _table: str, rowid: int) -> None:
        self.panes.delete(rowid)

    def on_select(self, _event: tk.Event) -> None:
        if selected := self.panes.selection():
            with self.conn:
                self.conn.execute('update selected_pane set pane_id = ?', (selected[-1], ))

    def add_pane(self) -> None:
        with undogroup(self.conn):
            self.conn.execute('''
                insert into pane (title, description, x, y, w, h) values (
                    'Pane '||coalesce((select max(id) + 1 from pane), 1),
                    '',
                    10,
                    10,
                    300,
                    200
                )
            ''')
        logger.info(f'Created new pane with id {self.conn.last_insert_rowid()}')

    def delete_selected(self) -> None:
        if selected := self.panes.selection():
            with undogroup(self.conn):
                for id in selected:
                    self.conn.execute('delete from pane where id = ?', (id, ))


class PaneViewer(ttk.Frame):

    def __init__(self, parent: tk.Misc, conn: apsw.Connection, listener: ChangeListener, id: int) -> None:
        super().__init__(parent, padding=5, relief='raised')
        self.conn = conn
        self.listener = listener
        self.id = id
        row = self.conn.execute('''
            select title, description, x, y, w, h from pane where id = ?
        ''', (self.id, )).fetchone()
        assert row is not None
        title, desc, self.x, self.y, self.w, self.h = row

        self.ghost_frame = ttk.Frame(parent, relief='solid')

        self.place(x=self.x, y=self.y, width=self.w, height=self.h)

        handle = tk.Canvas(self, width=10, height=0, cursor='fleur')
        handle.grid(row=0, column=0, sticky='ns')
        for y in range(0, 30, 5):
            handle.create_oval(0, y, 3, y + 3, fill='gray32', outline='')
            handle.create_oval(5, y, 8, y + 3, fill='gray32', outline='')

        def xy_drag(event: tk.Event) -> None:
            self.ghost_frame.place(width=self.w,
                                   height=self.h,
                                   x=self.x + event.x,
                                   y=self.y + event.y)

        def xy_release(event: tk.Event) -> None:
            self.ghost_frame.place_forget()
            self.x += event.x
            self.y += event.y
            with undogroup(self.conn):
                self.conn.execute('''
                    update pane set x = ?, y = ? where id = ?
                ''', (self.x, self.y, self.id))
            logger.info(f'Pane with id {self.id} dragged to ({self.x}, {self.y})')

        handle.bind('<B1-Motion>', xy_drag)
        handle.bind('<ButtonRelease-1>', xy_release)

        self.title_lbl = ttk.Label(self, text=title, font=('normal', 14), cursor='fleur')
        self.title_lbl.grid(row=0, column=1, sticky='ew')
        self.title_lbl.bind('<B1-Motion>', xy_drag)
        self.title_lbl.bind('<ButtonRelease-1>', xy_release)

        self.desc_lbl = ttk.Label(self, text=desc)
        self.desc_lbl.grid(row=1, column=0, columnspan=2, sticky='ew', pady=5)

        self.output = PaneOutput(self, self.conn, self.listener, self.id)
        self.output.grid(row=2, column=0, columnspan=2, sticky='nsew')

        self.rowconfigure(2, weight=1)
        self.columnconfigure(1, weight=1)

        grip = ttk.Sizegrip(self)
        grip.grid(row=3, column=2, sticky='se')

        def wh_click(event: tk.Event) -> str:
            return 'break'

        def wh_drag(event: tk.Event) -> str:
            self.ghost_frame.place(width=self.w + event.x,
                                   height=self.h + event.y,
                                   x=self.x,
                                   y=self.y)
            return 'break'

        def wh_release(event: tk.Event) -> str:
            self.ghost_frame.place_forget()
            self.w += event.x
            self.h += event.y
            with undogroup(self.conn):
                self.conn.execute('''
                    update pane set w = ?, h = ? where id = ?
                ''', (self.w, self.h, self.id))
            logger.info(f'Pane with id {id} resized to {self.w}x{self.h}')
            return 'break'

        grip.bind('<Button-1>', wh_click)
        grip.bind('<B1-Motion>', wh_drag)
        grip.bind('<ButtonRelease-1>', wh_release)

        self.listener.register(self.on_update, apsw.SQLITE_UPDATE, 'main', 'pane')

    @override
    def destroy(self) -> None:
        self.listener.unregister(self.on_update)
        super().destroy()

    def on_update(self, _type: int, _db: str, _table: str, rowid: int) -> None:
        row = self.conn.execute('''
            select title, description, x, y, w, h from pane where id = ?
        ''', (self.id, )).fetchone()
        assert row is not None
        title, desc, self.x, self.y, self.w, self.h = row
        self.title_lbl['text'] = title
        self.desc_lbl['text'] = desc
        self.place_configure(width=self.w, height=self.h, x=self.x, y=self.y)


class PaneOutput(ttk.Frame):
    def __init__(self, parent: tk.Misc, conn: apsw.Connection, listener: ChangeListener, id: int) -> None:
        super().__init__(parent, relief='raised')
        self.conn = conn
        self.listener = listener
        self.id = id
        self.output = runtime.run_pane(self.conn, self.id)
        self.render_output()
        self.listener.register(self.on_change, None, 'main', None, None)

    @override
    def destroy(self) -> None:
        self.listener.unregister(self.on_change)
        super().destroy()

    def on_change(self, _type: int, _db: str, table: str, _rowid: int) -> None:
        if self.conn.execute('select enabled from trigger_settings').fetchone() != (1, ):
            return # prevent cycles when nodes set an input value or control info
        # TODO: optimize this to only run when something actually changed
        self.output = runtime.run_pane(self.conn, self.id)
        self.render_output()

    def render_output(self) -> None:
        for child in self.winfo_children():
            child.destroy()

        if isinstance(self.output, bool) or isinstance(self.output, int) or isinstance(self.output, float) or isinstance(self.output, str):
            text = str(self.output)
            if isinstance(self.output, bool):
                text = _('True') if self.output else _('False')
            label = ttk.Label(self, text=text, font=('normal', 16), anchor='center')
            label.pack(side='top', fill='both', expand=True)
        elif isinstance(self.output, pd.DataFrame):
            DataFrameView(self, self.output).pack(side='top', fill='both', expand=True)
        elif isinstance(self.output, Figure):
            graph = FigureCanvasTkAgg(self.output, master=self) # type: ignore
            graph.draw() # type: ignore
            toolbar = NavigationToolbar2Tk(graph, self, pack_toolbar=False) # type: ignore
            toolbar.update()
            graph.mpl_connect('key_press_event', key_press_handler) # type: ignore
            graph.get_tk_widget().grid(row=0, column=0, sticky='nsew') # type: ignore
            toolbar.grid(row=1, column=0, sticky='ew')


class DataFrameView(ttk.Frame):
    def __init__(self, parent: tk.Misc, df: pd.DataFrame) -> None:
        super().__init__(parent)
        self.df = df

        self.table = ttk.Treeview(self)
        self.xscroll = ttk.Scrollbar(self, command=self.table.xview, orient='horizontal')
        self.yscroll = ttk.Scrollbar(self, command=self.table.yview, orient='vertical')
        self.table['xscrollcommand'] = self.xscroll.set
        self.table['yscrollcommand'] = self.yscroll.set

        self.xscroll.pack(side='bottom', fill='both')
        self.table.pack(side='left', fill='both', expand=True)
        self.yscroll.pack(side='left', fill='both')

        columns = [col if isinstance(col, str) else '.'.join(col) for col in self.df.columns.values]
        self.table['columns'] = columns
        for col in columns:
            self.table.column(col, width=100)
            self.table.heading(col, text=col)

        for idx, *row in df.itertuples(name=None):
            self.table.insert('', 'end', idx, text=str(idx), values=row)
            

class PaneSettings(ttk.Frame):
    def __init__(self, parent: tk.Misc, conn: apsw.Connection, listener: ChangeListener) -> None:
        super().__init__(parent, relief='sunken', padding=5)
        self.conn = conn
        self.listener = listener
        self.id: int | None = None
        self.listener.register(self.on_selected_update, apsw.SQLITE_UPDATE, 'main', 'selected_pane')
        self.setup()

    def setup(self) -> None:
        row = self.conn.execute('select pane_id from selected_pane').fetchone()
        assert row is not None
        id, = row
        assert id is None or isinstance(id, int)
        if id == self.id:
            return
        self.id = id

        for widget in self.winfo_children():
            widget.destroy()

        row = self.conn.execute('''
            select title, description, nodegroup_id, x, y, w, h from pane where id = ?
        ''', (self.id, )).fetchone()

        if self.id is None or row is None:
            ttk.Label(self, text=_('No pane selected')).grid(row=0, column=0, sticky='w')
            return

        title, desc, nodegroup_id, x, y, w, h = row
        self.title_var = tk.StringVar(value=title)
        self.desc_var = tk.StringVar(value=desc)
        self.x_var = tk.IntVar(value=x)
        self.y_var = tk.IntVar(value=y)
        self.w_var = tk.IntVar(value=w)
        self.h_var = tk.IntVar(value=h)
        self.title_var.trace_add('write', self.on_var_write)
        self.desc_var.trace_add('write', self.on_var_write)
        self.x_var.trace_add('write', self.on_var_write)
        self.y_var.trace_add('write', self.on_var_write)
        self.w_var.trace_add('write', self.on_var_write)
        self.h_var.trace_add('write', self.on_var_write)

        pane_lbl = ttk.Label(self, textvariable=self.title_var, font=('normal', 14))
        pane_lbl.grid(row=0, column=0, sticky='w')

        padding = (0, 3, 30)

        ttk.Label(self, text=_('Pane title:'), padding=padding).grid(row=1, column=0, sticky='w')
        title_entry = ttk.Entry(self, textvariable=self.title_var)
        title_entry.grid(row=1, column=1, sticky='ew')

        ttk.Label(self, text=_('Description:'), padding=padding).grid(row=2, column=0, sticky='w')
        desc_entry = ttk.Entry(self, textvariable=self.desc_var)
        desc_entry.grid(row=2, column=1, sticky='ew')

        ttk.Label(self, text=_('X position:'), padding=padding).grid(row=3, column=0, sticky='w')
        x_spin = ttk.Spinbox(self, textvariable=self.x_var, increment=1, from_=0, to=100000)
        x_spin.grid(row=3, column=1, sticky='ew')

        ttk.Label(self, text=_('Y position:'), padding=padding).grid(row=4, column=0, sticky='w')
        y_spin = ttk.Spinbox(self, textvariable=self.y_var, increment=1, from_=0, to=100000)
        y_spin.grid(row=4, column=1, sticky='ew')

        ttk.Label(self, text=_('Width:'), padding=padding).grid(row=5, column=0, sticky='w')
        w_spin = ttk.Spinbox(self, textvariable=self.w_var, increment=1, from_=0, to=100000)
        w_spin.grid(row=5, column=1, sticky='ew')

        ttk.Label(self, text=_('Height:'), padding=padding).grid(row=6, column=0, sticky='w')
        h_spin = ttk.Spinbox(self, textvariable=self.h_var, increment=1, from_=0, to=100000)
        h_spin.grid(row=6, column=1, sticky='ew')

        self.nodes_btn = ttk.Button(self, text=_('Create a node group for this pane'), command=self.create_nodegroup)
        if nodegroup_id is None:
            self.nodes_btn.grid(row=7, column=0, columnspan=2, sticky='ew', pady=20)

        self.listener.unregister(self.on_update)
        self.listener.unregister(self.on_delete)
        self.listener.register(self.on_update, apsw.SQLITE_UPDATE, 'main', 'pane', self.id)
        self.listener.register(self.on_delete, apsw.SQLITE_DELETE, 'main', 'pane', self.id)

    @override
    def destroy(self) -> None:
        self.listener.unregister(self.on_update)
        self.listener.unregister(self.on_delete)
        self.listener.unregister(self.on_selected_update)
        super().destroy()

    def on_update(self, _type: int, _db: str, _table: str, _rowid: int) -> None:
        row = self.conn.execute('''
            select title, description, nodegroup_id, x, y, w, h from pane where id = ?
        ''', (self.id, )).fetchone()
        assert row is not None
        title, desc, nodegroup_id, x, y, w, h = row
        self.title_var.set(title)
        self.desc_var.set(desc)
        self.x_var.set(x)
        self.y_var.set(y)
        self.w_var.set(w)
        self.h_var.set(h)
        if nodegroup_id is None:
            self.nodes_btn.grid(row=7, column=0, columnspan=2, sticky='ew', pady=20)
        else:
            self.nodes_btn.grid_forget()

    def on_delete(self, _type: int, _db: str, _table: str, _rowid: int) -> None:
        self.setup()

    def on_selected_update(self, _type: int, _db: str, _table: str, _rowid: int) -> None:
        self.setup()
        with self.conn:
            self.conn.execute('''
                update selected_nodegroup set nodegroup_id = (
                    select nodegroup_id
                        from pane
                        where id = (select pane_id from selected_pane)
                )
            ''')

    def on_var_write(self, _var: str, _index: str, _mode: str) -> None:
        try:
            title = self.title_var.get()
            desc = self.desc_var.get()
            x = self.x_var.get()
            y = self.y_var.get()
            w = self.w_var.get()
            h = self.h_var.get()
        except tk.TclError: # str -> int conversion might fail
            return

        row = title, desc, x, y, w, h

        # prevent infinite loop with change listener
        old_row = self.conn.execute('''
            select title, description, x, y, w, h from pane where id = ?
        ''', (self.id, )).fetchone()
        if row == old_row:
            return

        with undogroup(self.conn):
            self.conn.execute('''
                update pane set title = ?, description = ?, x = ?, y = ?, w = ?, h = ?
                where id = ?
            ''', (*row, self.id))

    def create_nodegroup(self) -> None:
        with undogroup(self.conn):
            self.conn.execute('insert into nodegroup (name) values (?)', (self.title_var.get(), ))
            id = self.conn.last_insert_rowid()
            self.conn.execute('update pane set nodegroup_id = ? where id = ?', (id, self.id))
            self.conn.execute('''
                insert into nodegroup_input (nodegroup_id, name, idx) values (:id, 'Data', 0);
                insert into nodegroup_output (nodegroup_id, name, idx) values (:id, 'Output', 0);
            ''', {'id': id})
            plugin.node_registry['Group Input'].create(self.conn, id)
            plugin.node_registry['Group Output'].create(self.conn, id)
        self.conn.execute('update selected_nodegroup set nodegroup_id = ?', (id, ))
