# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import contextlib
import io
import logging
import tempfile
import unittest
from pathlib import Path
from unittest import mock
from biomeca import main


class TestCli(unittest.TestCase):
    def test_version_flag(self):
        version = None
        with open(Path(__file__).parent.parent / 'pyproject.toml') as file:
            for line in file:
                if line.startswith('version ='):
                    version = line.split('"')[1]
                    break

        for flag in ('-V', '--version'):
            with self.subTest(flag=flag):
                stdout = io.StringIO()
                with contextlib.redirect_stdout(stdout):
                    main.main(['biomeca', flag])
                self.assertEqual(stdout.getvalue().rstrip(), f'biomeca v{version}')

    def test_file_arg(self):
        with tempfile.NamedTemporaryFile(suffix='.sqlite', delete_on_close=False) as tmp:
            path = tmp.name
            tmp.close()
            with mock.patch('biomeca.widgets.MainWindow') as MockWindow:
                main.main(['biomeca', '--', path])
                MockWindow.assert_called_with(path=path)


class TestCliVerboseFlag(unittest.TestCase):
    def setUp(self):
        self.root = logging.getLogger()
        self.root.setLevel(logging.NOTSET)
        self.patcher = mock.patch('biomeca.widgets.MainWindow')
        self.patcher.start()

    def tearDown(self):
        self.patcher.stop()

    def test_verbose_flag_warning(self):
        with contextlib.redirect_stderr(io.StringIO()):
            main.main(['biomeca'])
        self.assertEqual(self.root.level, logging.WARNING)

    def test_verbose_flag_info(self):
        with contextlib.redirect_stderr(io.StringIO()):
            main.main(['biomeca', '-v'])
        self.assertEqual(self.root.level, logging.INFO)

    def test_verbose_flag_debug(self):
        with contextlib.redirect_stderr(io.StringIO()):
            main.main(['biomeca', '-vv'])
        self.assertEqual(self.root.level, logging.DEBUG)
