# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import sqlite3
import tempfile
import tkinter as tk
import unittest
from tkinter import ttk
from unittest import mock
from biomeca import widgets
from widget_utils import get_all_by_option, get_by_option, get_menuitem_by_option, pump_events, parent


class TestQuit(unittest.TestCase):
    def setUp(self):
        self.patcher = mock.patch.object(tk.Tk, 'quit')
        self.quit = self.patcher.start()
        self.root = tk.Tk()
        self.win = widgets.MainWindow(self.root)
        pump_events(self.root)

    def tearDown(self):
        self.patcher.stop()
        self.root.destroy()
        pump_events(self.root)

    def test_quit_button(self):
        file_menu = get_by_option(self.root, tk.Menu, label='File')
        self.assertIsNotNone(file_menu)
        quit_idx = get_menuitem_by_option(file_menu, label='Quit')
        self.assertIsNotNone(quit_idx)
        file_menu.invoke(quit_idx)
        pump_events(self.root)
        self.quit.assert_called()


class TestProjectFileManagement(unittest.TestCase):
    def setUp(self):
        self.root = tk.Tk()
        self.win = widgets.MainWindow(self.root)
        pump_events(self.root)

    def tearDown(self):
        self.root.destroy()
        pump_events(self.root)

    def test_unsaved_on_startup(self):
        self.assertEqual(self.root.wm_title(), 'Unsaved')

    @mock.patch('tkinter.filedialog.askopenfilename')
    @mock.patch('tkinter.filedialog.asksaveasfilename')
    def test_open_valid_project_menubuttons(self, asksaveas, askopen):
        file_menu = get_by_option(self.root, tk.Menu, label='File')
        self.assertIsNotNone(file_menu)
        new_idx = get_menuitem_by_option(file_menu, label='New Project')
        open_idx = get_menuitem_by_option(file_menu, label='Open Project')
        save_as_idx = get_menuitem_by_option(file_menu, label='Save As')
        self.assertIsNotNone(new_idx)
        self.assertIsNotNone(open_idx)
        self.assertIsNotNone(save_as_idx)

        with tempfile.NamedTemporaryFile(suffix='.sqlite', delete_on_close=False) as tmp:
            path = tmp.name
            tmp.close()

            asksaveas.return_value = path
            file_menu.invoke(save_as_idx)
            pump_events(self.root)
            asksaveas.assert_called()
            self.assertEqual(self.root.wm_title(), 'Unsaved')

            askopen.return_value = path
            file_menu.invoke(open_idx)
            pump_events(self.root)
            askopen.assert_called()
            self.assertEqual(self.root.wm_title(), path)

            # close the file before context manager exit
            file_menu.invoke(new_idx)
            self.assertEqual(self.root.wm_title(), 'Unsaved')


class TestPanesList(unittest.TestCase):
    def setUp(self):
        self.root = tk.Tk()
        self.win = widgets.MainWindow(self.root)
        pump_events(self.root)
        self.setup_widgets()

    def tearDown(self):
        pump_events(self.root)
        self.root.destroy()
        pump_events(self.root)

    def setup_widgets(self):
        self.add_btn = get_by_option(self.root, ttk.Button, text='Add')
        self.delete_btn = get_by_option(self.root, ttk.Button, text='Delete')
        self.assertIsNotNone(self.add_btn)
        self.assertIsNotNone(self.delete_btn)
        self.tree = get_by_option(parent(self.root, self.add_btn, 2), ttk.Treeview)
        self.assertIsNotNone(self.tree)

        self.edit_menu = get_by_option(self.root, tk.Menu, label='Edit')
        self.assertIsNotNone(self.edit_menu)
        self.undo_idx = get_menuitem_by_option(self.edit_menu, label='Undo')
        self.redo_idx = get_menuitem_by_option(self.edit_menu, label='Redo')
        self.assertIsNotNone(self.undo_idx)
        self.assertIsNotNone(self.redo_idx)

    def test_list_add_button(self):
        self.add_btn.invoke()
        pump_events(self.root)
        self.assertEqual(len(self.tree.get_children('')), 1)

        self.add_btn.invoke()
        pump_events(self.root)
        self.assertEqual(len(self.tree.get_children('')), 2)

    def test_list_delete_button(self):
        self.add_btn.invoke()
        pump_events(self.root)
        children = self.tree.get_children('')
        self.assertEqual(len(children), 1)

        self.tree.selection_set(*children)
        pump_events(self.root)
        self.delete_btn.invoke()
        pump_events(self.root)
        self.assertEqual(len(self.tree.get_children('')), 0)

    def test_list_delete_multiple(self):
        for _ in range(5):
            self.add_btn.invoke()
            pump_events(self.root)
        children = self.tree.get_children('')
        self.assertEqual(len(children), 5)

        self.tree.selection_set(children[0], children[1], children[3])
        pump_events(self.root)
        self.delete_btn.invoke()
        pump_events(self.root)
        new_children = self.tree.get_children('')
        self.assertEqual(len(new_children), 2)
        self.assertCountEqual(new_children, [children[2], children[4]])

        self.tree.selection_set(children[2], children[4])
        pump_events(self.root)
        self.delete_btn.invoke()
        pump_events(self.root)
        self.assertEqual(len(self.tree.get_children('')), 0)

    def test_list_undo_add_delete(self):
        for _ in range(2):
            self.add_btn.invoke()
            pump_events(self.root)
        children = self.tree.get_children('')
        self.assertEqual(len(children), 2)

        self.tree.selection_set(*children)
        pump_events(self.root)
        self.delete_btn.invoke()
        pump_events(self.root)
        self.assertEqual(len(self.tree.get_children('')), 0)

        actions = [
            (self.undo_idx, 2),
            (self.undo_idx, 1),
            (self.undo_idx, 0),
            (self.undo_idx, 0),
            (self.redo_idx, 1),
            (self.redo_idx, 2),
            (self.redo_idx, 0),
            (self.redo_idx, 0),
        ]

        for i, (idx, count) in enumerate(actions):
            action = 'undo' if idx == self.undo_idx else 'redo'
            with self.subTest(i=i, action=action, expected_count=count):
                self.edit_menu.invoke(idx)
                pump_events(self.root)
                self.assertEqual(len(self.tree.get_children('')), count)

    @mock.patch('tkinter.filedialog.askopenfilename')
    @mock.patch('tkinter.filedialog.asksaveasfilename')
    def test_list_change_persistence(self, asksaveas, askopen):
        file_menu = get_by_option(self.root, tk.Menu, label='File')
        self.assertIsNotNone(file_menu)
        new_idx = get_menuitem_by_option(file_menu, label='New Project')
        open_idx = get_menuitem_by_option(file_menu, label='Open Project')
        save_as_idx = get_menuitem_by_option(file_menu, label='Save As')
        self.assertIsNotNone(new_idx)
        self.assertIsNotNone(open_idx)
        self.assertIsNotNone(save_as_idx)

        with tempfile.NamedTemporaryFile(suffix='.sqlite', delete_on_close=False) as tmp:
            path = tmp.name
            tmp.close()

            for _ in range(2):
                self.add_btn.invoke()
                pump_events(self.root)
            self.assertEqual(len(self.tree.get_children('')), 2)

            asksaveas.return_value = path
            file_menu.invoke(save_as_idx)
            pump_events(self.root)
            asksaveas.assert_called()
            self.assertEqual(self.root.wm_title(), 'Unsaved')

            file_menu.invoke(new_idx)
            self.assertEqual(self.root.wm_title(), 'Unsaved')
            self.setup_widgets()
            self.assertEqual(len(self.tree.get_children('')), 0)

            askopen.return_value = path
            file_menu.invoke(open_idx)
            pump_events(self.root)
            askopen.assert_called()
            self.assertEqual(self.root.wm_title(), path)
            self.setup_widgets()
            children = self.tree.get_children('')
            self.assertEqual(len(children), 2)
            self.tree.selection_set(*children)
            pump_events(self.root)
            self.delete_btn.invoke()
            pump_events(self.root)
            self.assertEqual(len(self.tree.get_children('')), 0)

            file_menu.invoke(new_idx)
            self.assertEqual(self.root.wm_title(), 'Unsaved')
            self.setup_widgets()
            self.assertEqual(len(self.tree.get_children('')), 0)

            askopen.return_value = path
            file_menu.invoke(open_idx)
            pump_events(self.root)
            askopen.assert_called()
            self.assertEqual(self.root.wm_title(), path)
            self.setup_widgets()
            self.assertEqual(len(self.tree.get_children('')), 0)

            # close the file before context manager exit
            file_menu.invoke(new_idx)
            self.assertEqual(self.root.wm_title(), 'Unsaved')


class TestPaneSettings(unittest.TestCase):
    def setUp(self):
        self.root = tk.Tk()
        self.win = widgets.MainWindow(self.root)
        pump_events(self.root)
        self.setup_widgets()

    def tearDown(self):
        pump_events(self.root)
        self.root.destroy()
        pump_events(self.root)

    def setup_widgets(self):
        self.edit_menu = get_by_option(self.root, tk.Menu, label='Edit')
        self.assertIsNotNone(self.edit_menu)
        self.undo_idx = get_menuitem_by_option(self.edit_menu, label='Undo')
        self.redo_idx = get_menuitem_by_option(self.edit_menu, label='Redo')
        self.assertIsNotNone(self.undo_idx)
        self.assertIsNotNone(self.redo_idx)

        self.add_btn = get_by_option(self.root, ttk.Button, text='Add')
        self.assertIsNotNone(self.add_btn)
        self.tree = get_by_option(parent(self.root, self.add_btn, 2), ttk.Treeview)
        self.assertIsNotNone(self.tree)

        self.add_btn.invoke()
        pump_events(self.root)
        children = self.tree.get_children('')
        self.assertEqual(len(children), 1)
        self.tree.selection_set(*children)
        pump_events(self.root)

        title_label = get_by_option(self.root, ttk.Label, text='Pane title:')
        desc_label = get_by_option(self.root, ttk.Label, text='Description:')
        x_label = get_by_option(self.root, ttk.Label, text='X position:')
        y_label = get_by_option(self.root, ttk.Label, text='Y position:')
        w_label = get_by_option(self.root, ttk.Label, text='Width:')
        h_label = get_by_option(self.root, ttk.Label, text='Height:')
        self.assertIsNotNone(title_label)
        self.assertIsNotNone(desc_label)
        self.assertIsNotNone(x_label)
        self.assertIsNotNone(y_label)
        self.assertIsNotNone(w_label)
        self.assertIsNotNone(h_label)
        self.settings = parent(self.root, title_label)
        entries = [w for w in self.settings.grid_slaves() if isinstance(w, ttk.Entry)]
        row = lambda w: w.grid_info()['row']
        self.title_entry = next((w for w in entries if row(w) == row(title_label)), None)
        self.desc_entry = next((w for w in entries if row(w) == row(desc_label)), None)
        self.x_entry = next((w for w in entries if row(w) == row(x_label)), None)
        self.y_entry = next((w for w in entries if row(w) == row(y_label)), None)
        self.w_entry = next((w for w in entries if row(w) == row(w_label)), None)
        self.h_entry = next((w for w in entries if row(w) == row(h_label)), None)
        self.assertIsNotNone(self.title_entry)
        self.assertIsNotNone(self.desc_entry)
        self.assertIsNotNone(self.x_entry)
        self.assertIsNotNone(self.y_entry)
        self.assertIsNotNone(self.w_entry)
        self.assertIsNotNone(self.h_entry)

        self.title_var = tk.StringVar(name=self.title_entry['textvariable'])
        self.x_var = tk.IntVar(name=self.x_entry['textvariable'])
        self.y_var = tk.IntVar(name=self.y_entry['textvariable'])
        self.w_var = tk.IntVar(name=self.w_entry['textvariable'])
        self.h_var = tk.IntVar(name=self.h_entry['textvariable'])

        self.pane_viewer = get_by_option(self.root, widgets.PaneViewer)
        self.assertIsNotNone(self.pane_viewer)
        self.pane_viewer_title = get_by_option(self.pane_viewer, ttk.Label, text='Pane 1')
        self.assertIsNotNone(self.pane_viewer_title)

    def test_title_change_and_undo(self):
        titles = ['Foo Pane', 'a', '', 'Pane 4294967296', '123', '\');--']
        for title in titles:
            with self.subTest(title=title):
                old_title = self.title_var.get()
                self.title_var.set(title)
                pump_events(self.root)
                self.assertEqual(self.pane_viewer_title['text'], title)

                self.edit_menu.invoke(self.undo_idx)
                pump_events(self.root)
                self.assertEqual(self.pane_viewer_title['text'], old_title)

                self.edit_menu.invoke(self.redo_idx)
                pump_events(self.root)
                self.assertEqual(self.pane_viewer_title['text'], title)

    def test_position_change(self):
        old_x = self.x_var.get()
        old_y = self.y_var.get()
        x = 42
        y = 123

        self.assertEqual(int(self.pane_viewer.place_info()['x']), old_x)
        self.assertEqual(int(self.pane_viewer.place_info()['y']), old_y)

        self.x_var.set(x)
        self.y_var.set(y)
        pump_events(self.root)
        self.assertEqual(int(self.pane_viewer.place_info()['x']), x)
        self.assertEqual(int(self.pane_viewer.place_info()['y']), y)

    @unittest.expectedFailure
    def test_position_undo(self):
        old_x = self.x_var.get()
        old_y = self.y_var.get()
        x = 42
        y = 123

        self.x_var.set(x)
        self.y_var.set(y)
        pump_events(self.root)

        self.edit_menu.invoke(self.undo_idx)
        pump_events(self.root)
        self.assertEqual(int(self.pane_viewer.place_info()['x']), old_x)
        self.assertEqual(int(self.pane_viewer.place_info()['y']), old_y)

        self.edit_menu.invoke(self.redo_idx)
        pump_events(self.root)
        self.assertEqual(int(self.pane_viewer.place_info()['x']), x)
        self.assertEqual(int(self.pane_viewer.place_info()['y']), y)

    def test_size_change(self):
        old_w = self.w_var.get()
        old_h = self.h_var.get()
        w = 400
        h = 300

        self.assertEqual(int(self.pane_viewer.place_info()['width']), old_w)
        self.assertEqual(int(self.pane_viewer.place_info()['height']), old_h)

        self.w_var.set(w)
        self.h_var.set(h)
        pump_events(self.root)
        self.assertEqual(int(self.pane_viewer.place_info()['width']), w)
        self.assertEqual(int(self.pane_viewer.place_info()['height']), h)

    @unittest.expectedFailure
    def test_size_undo(self):
        old_w = self.w_var.get()
        old_h = self.h_var.get()
        w = 400
        h = 300

        self.w_var.set(w)
        self.h_var.set(h)
        pump_events(self.root)

        self.edit_menu.invoke(self.undo_idx)
        pump_events(self.root)
        self.assertEqual(int(self.pane_viewer.place_info()['width']), old_w)
        self.assertEqual(int(self.pane_viewer.place_info()['height']), old_h)

        self.edit_menu.invoke(self.redo_idx)
        pump_events(self.root)
        self.assertEqual(int(self.pane_viewer.place_info()['width']), w)
        self.assertEqual(int(self.pane_viewer.place_info()['height']), h)

    @unittest.expectedFailure
    def test_create_nodegroup(self):
        create_btn = get_by_option(self.root, ttk.Button, text='Create a node group for this pane')
        self.assertIsNotNone(create_btn)
        create_btn.invoke()
        pump_events(self.root)
        create_btn = get_by_option(self.root, ttk.Button, text='Create a node group for this pane')
        self.assertIsNone(create_btn)

        self.edit_menu.invoke(self.undo_idx)
        pump_events(self.root)
        create_btn = get_by_option(self.root, ttk.Button, text='Create a node group for this pane')
        self.assertIsNotNone(create_btn)

        self.edit_menu.invoke(self.redo_idx)
        pump_events(self.root)
        create_btn = get_by_option(self.root, ttk.Button, text='Create a node group for this pane')
        self.assertIsNone(create_btn)
