# SPDX-FileCopyrightText: 2024 Sam Nystrom <sam@samnystrom.dev>
# SPDX-License-Identifier: GPL-3.0-only

import tkinter as tk


def get_by_option(parent: tk.Misc, cls: type[tk.Misc] | None = None, **options: str) -> tk.Misc | None:
    matches = get_all_by_option(parent, cls, **options)
    return matches[0] if len(matches) > 0 else None


def get_all_by_option(parent: tk.Misc, cls: type[tk.Misc] | None = None, **options: str) -> list[tk.Misc]:
    '''Recursively traverse child widgets and menu items of parent to find a widget of type cls that matches all of the given options.
    '''
    results = []
    for option, value in options.items():
        try:
            if parent.cget(option) != value:
                break
        except:
            break
    else:
        if cls is None or isinstance(parent, cls):
            results.append(parent)

    for child in parent.winfo_children():
        if widgets := get_all_by_option(child, cls, **options):
            results += widgets

    if isinstance(parent, tk.Menu):
        widget_idx = -1
        for i in range(0, parent.index('end') + 1): # type: ignore
            matches = False
            for option, value in options.items():
                try:
                    if parent.entrycget(i, option) != value:
                        break
                except:
                    break
            else:
                matches = True

            if parent.type(i) == 'cascade':
                widget_idx += 1
                child = parent.winfo_children()[widget_idx]
                if matches and (cls is None or isinstance(child, cls)):
                    results.append(child)
                if widgets := get_all_by_option(child, cls, **options):
                    results += widgets

    return results


def get_menuitem_by_option(menu: tk.Menu, **options: str) -> int | None:
    '''Return the index of the first item in menu that matches all of the given options, or None if no items match.
    '''
    for i in range(0, menu.index('end') + 1): # type: ignore
        for option, value in options.items():
            try:
                if menu.entrycget(i, option) != value:
                    break
            except:
                break
        else:
            return i
    return None


def pump_events(root: tk.Tk) -> None:
    while root.dooneevent(tk._tkinter.ALL_EVENTS | tk._tkinter.DONT_WAIT):
        pass
    root.update()


def parent(root: tk.Tk, widget: tk.Misc, n: int = 1) -> tk.Misc:
    for _ in range(n):
        widget = root.nametowidget(widget.winfo_parent())
    return widget
